/****************************************************************==
** GD extension for TCL. 
** The focus is on output only.
** For now the model is based on the function calls in GD.
** Only a modicum of effort is put into encapsulating the
** commands in an object mechanism.
** GIF files are disabled in this version, though they may
** be enabled in the future if solution can be found to the
** licensing problem.
** PNG is the default format.
** JPEG is also supported.
**
** Build environment:
** 
** Requires: (Versions used for development in parentheses)
**           PNG (1.0.8), GD (2.0.1 + extensions), ZLIB (1.1.3)
** Optional: (Versions used for development in parentheses)
**           JPEG (6b), freetype (2.0.4)
** 
** Purpose: Writing JPEG and PNG images from Tcl programs.
**          Does not require Tk.
**          GIF support has been removed
** 
** Model: Currently follows the model for GDI, an experimental extension 
** for use in driving printers. That model, in turn, is based on the Tk 
** canvas model. As with GDI, more options are documented than supported.
** I like the idea of keeping these two extensions in sync, since the goals 
** are similar--non-interactive drawing output.
** 
** New or different: 
** o Has NO RELATION to a different product wrapping GD for tcl called
**   'gdtcl'
** o Emits palette-based PNG images.
** o Reads palette-based PNG images (very limited at this time).
** o Emits ellipses properly.
** o Uses ellipse code to generate arcs
** o Supports thicknesses for geometric shapes
** o Supports Truetype fonts and rotation of text.
** o No longer supports GIF images
** o Does NOT support XPM images at this time.
** 
** Usage:
** gd [open|close|config|fill|arc|line|oval|polygon|rectangle|text|fontalias|version]
** 
** General Remarks:
**   Color is specified in any of several ways:
**    - An integer triplet (list)
**    - An octothorp (#) followed by a 6-digit hex value
**       The integers are interpreted as RGB and a color is allocated in 
**       the palette if necessary. Color indices are always reused if 
**       possible. If no colors remain to allocate, the closest available 
**       color is used.
**    - An equal sign (=) followed by a list of integers
**       The integers are interpreted as GD's "styled line" specification.
**    - An ampersand (&) followed by a single integer
**       The integer is interpreted as an existing color index
**    - An at sign (@) followed by a valid image name
**       The image name is used as a GD "brush"
**    - A percent sign (%) followed by a valid image name
**       The image name is used as a GD "tile"
** 
** Image names are created and returned solely by the "gd open" command and may be used 
** until the "gd close" command operates on them. Image names are validated 
** whenever used in a gd command. An error results if the image name is not valid.
** 
** gd open
**   gd open -width w -height h -file f -filetype [png|xbm] -background color
**     Creates a new gd image and returns a token for it.
**     The width and height of the image are established at this time and 
**     cannot be changed.
**     The width and height are either explicitly set, or are set to the 
**     width and height of the image mentioned in the -file argument.
**     If both width and height and file are specified, the width and 
**     height of the image are as specified, and the file is scaled to fit 
**     into that image.
**     If the file name ends with .png, it is assumed to be a PNG image.
**     If the file name ends with .xbm, it is assumed to be an XBM image.
**     Use of the -filetype argument explicitly documents which kind of 
**     image is present.
**     The -background is used to set the background color only on images 
**     that were not initialized from files.
**   LIMITATION: Currently only 8-bit paletted PNG images are supported.
**     No checks are currently made on the image type read, so reading a truecolor
**     image will likely cause a core dump. This will be fixed as I 
**     understand PNG better and as the "real" design of GD2.0 takes shape.
**     This version uses the GD1.2 code for GIFs instead of the GD1.3 code 
**     for GIFs.
** 
** gd close
**   gd close image [-save [-file filename [-filetype [gif|png] ] ] ]
**     Closing an image will make it invalid for future use. Possibly a 
**     'gd save' command may be implemented to allow intermediate results 
**     to be written to a file.
**     If -save is not specified, the image is discarded entirely.
**     If -file is present, -save is assumed; if -file is not present,
**     the filename is 'image.png', and is in PNG format.
**     If -filetype is present, that filetype is explicitly used, 
**     irrespective of file name.
** 
** gd config
**   gd config image [-brush image] [-tile  image] [-style int_list] 
**        [-interlace 0|1] [-transparent color] [-colormap] [-size] 
**        [-fonttable arrayname]
**     The config command is used to access and set many 'image state variables'
**     With no arguments, it provides information about all items that 
**     represent the configuration.
**     With an argument and no value, that setting is reported in the 
**     return value.
**     With an argument and a value, that setting is changed for the image.
** 
** gd fill
**   gd fill image x y [-outline color] -fill color
**     Fills the given image starting at the given point with the given 
**     fill color.
**     The extension fills in both ways permitted by GD:
**     o If -outline is specified, fills until that color is reached (watch 
**       for leaks!)
**     o If -outline is not specified, fills anything matching the color at 
**       (x, y), starting at (x, y).
**   LIMITATION: Currently the fill algorithm misbehaves badly when a brush 
**   or tile is used as the fill color. Therefore, brushes and tiles are not 
**   permitted as fill colors this time.
**       
** gd arc
**   gd arc image x1 y1 x2 y2 -extent degrees 
** 	-fill color -outline color -outlinestipple bitmap 
**         -start degrees -stipple bitmap -style [pieslice|chord|arc] -width linewid
**     Draws an arc in the image along the orthorectified ellipse 
**     circumscribed by (x1,y1) and (x2,y2). Starts at the angle in -start, 
**     and extends for the amount specified in -extent.
**     Drawn with the -outline color (or styled line), drawn in the 
**     indicated style (arc is the default), and filled with the indicated color (if 
**     any). If filled, an 'arc' style is replaced with 'chord'.
** 
**   LIMITATION: -outlinestipple is not supported at this time.
**     Also, the mechanism for drawing wide arcs may need tweaking.
** 
** gd line
**   gd line image x1 y1 ... xn yn
**          -arrow [first|last|both|none] -arrowshape {d1 d2 d3} 
**          -capstyle [butt|projecting|round] -fill color 
**          -joinstyle [bevel|miter|round] -smooth [true|false] 
**          -splinesteps number -stipple bitmap -width linewid
** 
**     Draws an open polygonal line connecting the points in the indicated 
**     order. Uses the specified color (or styled line or brush) to draw 
**     the line.
** 
**   LIMITATION: -capstyle, -joinstyle, -smooth, 
**               -splinesteps, and -stipple are not supported at 
** 	      this time. [however, using various brushes and styles
** 	      for the outline can emulate several of these effects]
** gd oval
**   gd oval image x1 y1 x2 y2 -fill color -outline color
**          -stipple bitmap -width linewid
**     Draws an orthorectified ellipse circumscribed by (x1,y1) and (x2, y2).
**     Uses the indicated outline and fill colors.
** 
**   LIMITATION: -stipple is not supported at this time.
**               Since GD does not support filled ellipses at this time, 
** 	      filling is emulated with the FillToOutline function, with 
** 	      the limitations noted above. In particular, the fill will 
** 	      not work properly if there are embedded shapes in the oval 
** 	      with the same outline color as the oval.
** 	      Creating a reasonable filled oval is not difficult given the
** 	      Bresenham algorithm, so this should be fixed soon.
** 
** gd polygon
**   gd polygon image x1 y1 ... xn yn 
**          -fill color -outline color -smooth [true|false] 
** 	 -splinesteps number -stipple bitmap -width linewid
**     Creates a closed polygon with the given vertices, outline, and fill.
** 
**   LIMITATION: -smooth, -splitsteps, and -stipple are not 
**               supported at this time.
** 
** gd rectangle
**   gd rectangle image x1 y1 x2 y2 -fill color -outline color 
** 	 -stipple bitmap -width linewid
**     Creates a rectangle with corners at the given points with the given 
**     outline and fill.
** 
**   LIMITATION: -stipple is not supported at this time.
** 
** gd text
**   gd text image x y -anchor [center|n|e|s|w] -fill color -font fontname 
**          -justify  [left|right|center]  -stipple  bitmap  -text string  -width linelen
**          -anchor [nw|n|ne|w|center|e|sw|s|se]
** 	 -orientation horizontal|vertical
**     Creates a single line of text with the upper left corner of the 
**     first letter at the indiated point with the indicated fill color.
**     Orients text vertically or horizontally as requested.
**     Uses any of the 5 GD fonts: tiny, small, mediumbold, large, and 
**     giant. Default is mediumbold.
** 
**   LIMITATION: -stipple is not supported at this time.
**
** gd fontalias
**   gd fontalias globspec ...
**     Searches the files matching the specifications to extract font names from those files
**     This, in turn, supports building an aliasing mechanism to suggest which files be used for which
**     fonts.
**     Sample usage
**          set Freetype_Font_Aliases [ gd fontalias [ glob $env(GDFONTPATH)/*.ttf ] ]
**          ...
**          gd configure $im -fonttable Freetype_Font_Aliases
**     (env is a global variable, and im represents an image tag)
**
** gd version
**   Prints the current version number of this gd extension (0.3.0.3)
** 
** Copyright/license notices:
** 
** ** LIBPNG-1.0.3 **
** Copyright (c) 1995, 1996 Guy Eric Schalnat, Group 42, Inc.
** Copyright (c) 1996, 1997 Andreas Dilger
** Copyright (c) 1998, 1999 Glenn Randers-Pehrson
** 
** The PNG Reference Library (libpng) is supplied "AS IS".  The Contributing
** Authors and Group 42, Inc. disclaim all warranties, expressed or implied,
** including, without limitation, the warranties of merchantability and of
** fitness for any purpose.  The Contributing Authors and Group 42, Inc.
** assume no liability for direct, indirect, incidental, special, exemplary,
** or consequential damages, which may result from the use of the PNG
** Reference Library, even if advised of the possibility of such damage.
** 
** Permission is hereby granted to use, copy, modify, and distribute this
** source code, or portions hereof, for any purpose, without fee, subject
** to the following restrictions:
** 
**  1. The origin of this source code must not be
**     misrepresented.
** 
**  2. Altered versions must be plainly marked as such
**     and must not be misrepresented as being the
**     original source.
** 
**  3. This Copyright notice may not be removed or
**     altered from any source or altered source
**     distribution.
** 
** The Contributing Authors and Group 42, Inc. specifically permit, without
** fee, and encourage the use of this source code as a component to
** supporting the PNG file format in commercial products.  If you use this
** source code in a product, acknowledgment is not required but would be
** appreciated.
** 
** ** ZLIB 1.1.3 **
** 
**  (C) 1995-1998 Jean-loup Gailly and Mark Adler
** 
**   This software is provided 'as-is', without any express or implied
**   warranty.  In no event will the authors be held liable for any damages
**   arising from the use of this software.
** 
**   Permission is granted to anyone to use this software for any purpose,
**   including commercial applications, and to alter it and redistribute it
**   freely, subject to the following restrictions:
** 
**   1. The origin of this software must not be misrepresented; you must not
**      claim that you wrote the original software. If you use this software
**      in a product, an acknowledgment in the product documentation would be
**      appreciated but is not required.
**   2. Altered source versions must be plainly marked as such, and must not be
**      misrepresented as being the original software.
**   3. This notice may not be removed or altered from any source distribution.
** 
**   Jean-loup Gailly        Mark Adler
**   jloup@gzip.org          madler@alumni.caltech.edu
** 
** If you use the zlib library in a product, we would appreciate *not*
** receiving lengthy legal documents to sign. The sources are provided
** for free but without warranty of any kind.  The library has been
** entirely written by Jean-loup Gailly and Mark Adler; it does not
** include third-party code.
** 
** If you redistribute modified sources, we would appreciate that you include
** in the file ChangeLog history information documenting your changes.
** 
** ** GD **
** 
**      Portions copyright 1994, 1995, 1996, 1997, 1998, 1999, by Cold
**      Spring Harbor Laboratory. Funded under Grant P41-RR02188 by the
**      National Institutes of Health.
**      
**      Portions copyright 1996, 1997, 1998, 1999, by Boutell.Com, Inc.
**      
**      Portions relating to GD2 format copyright 1999 Philip Warner.
**      
**      Portions relating to PNG copyright 1999, Greg Roelofs.
**      
**      Portions relating to libttf copyright 1999, John Ellson
**      (ellson@lucent.com).
**      
**      _Permission has been granted to copy and distribute gd in any
**      context without fee, including a commercial application, provided
**      that this notice is present in user-accessible supporting
**      documentation._
**      
**      This does not affect your ownership of the derived work itself, and
**      the intent is to assure proper credit for the authors of gd, not to
**      interfere with your productive use of gd. If you have questions,
**      ask. "Derived works" includes all programs that utilize the
**      library. Credit must be given in user-accessible documentation.
**      
**      _This software is provided "AS IS."_ The copyright holders disclaim
**      all warranties, either express or implied, including but not
**      limited to implied warranties of merchantability and fitness for a
**      particular purpose, with respect to this code and accompanying
**      documentation.
**      
**      Although their code does not appear in gd 1.6.3, the authors wish
**      to thank David Koblas, David Rowley, and Hutchison Avenue Software
**      Corporation for their prior contributions.
** 
** *** GD 1.7.3 Modifications
** 
**      GD 1.7.3 modifications copyright Michael I. Schwartz, 1999.
** 
**      Permission is granted to copy and distribute these modifications
**      without fee, provided this copyright notice is maintained.
**      Software modifications are provided AS IS, with no express or implied
**      warranty of any kind.
** 
** ** FREETYPE 1.2
** 
** 
**     Copyright 1996 David Turner    <turner@enst.fr>
**             - 1998 Robert Wilhelm  <robert@physiol.med.tu-muenchen.de>
**                    Werner Lemberg  <wl@gnu.org>
** 
**   The FreeType  project is copyright (C) 1996-1998  by David Turner,
**   Robert Wilhelm, and Werner Lemberg.  All rights reserved except as
**   specified below.
** 
** 
** 1. No Warranty:
** 
**   THE FREETYPE ARCHIVE  IS PROVIDED "AS IS" WITHOUT  WARRANTY OF ANY
**   KIND, EITHER EXPRESSED OR  IMPLIED, INCLUDING, BUT NOT LIMITED TO,
**   WARRANTIES  OF  MERCHANTABILITY   AND  FITNESS  FOR  A  PARTICULAR
**   PURPOSE.  IN NO EVENT WILL ANY OF THE AUTHORS OR COPYRIGHT HOLDERS
**   BE LIABLE  FOR ANY DAMAGES CAUSED  BY THE USE OR  THE INABILITY TO
**   USE, OF THE FREETYPE PROJECT.
** 
**   As  you have  not signed  this license,  you are  not  required to
**   accept  it.   However,  as  the FreeType  project  is  copyrighted
**   material, only  this license, or  another one contracted  with the
**   authors, grants you  the right to use, distribute,  and modify it.
**   Therefore,  by  using,  distributing,  or modifying  the  FreeType
**   project, you indicate that you understand and accept all the terms
**   of this license.
** 
** 
** 2. Redistribution:
** 
**   Redistribution and use in source and binary forms, with or without
**   modification, are permitted provided that the following conditions
**   are met:
** 
**     o Redistribution  of source code  must retain this  license file
**       (LICENSE.TXT) unaltered;  any additions, deletions  or changes
**       to   the  original   files  must   be  clearly   indicated  in
**       accompanying  documentation.  All  copyright  notices must  be
**       preserved in all copies of source files.
** 
**     o Redistribution  in binary form must provide  a disclaimer that
**       states that the  software is based in part of  the work of the
**       FreeType  team  in the  distribution  documentation.  We  also
**       encourage you to  put an URL to the FreeType  web page in your
**       documentation, though this isn't mandatory.
** 
**   These conditions  apply to any  software derived from or  based on
**   the FreeType code, not just  the unmodified files.  If you use our
**   work, you  must acknowledge us.  However,  no fee need  be paid to
**   us.
** 
** ** GD TCL EXTENSION **
**         Written by Michael I. Schwartz
**         Copyright 1997, 1999 Schwartz Computer Consulting Services
** 	Permission granted to use this code in any fashion without fee,
**         provided that this notice is retained and any alterations are
** 	labeled as such.
**         The GD TCL extension is provided AS IS, with no express or implied
**         warranty of any kind.
** 	In no event will the author or copyright holders be
** 	liable for any damages caused by the use or inability to use
** 	the GD TCL extension.
** 
**
** Tcl Extension for Windows
** RCS Version $Revision: 1.22 $
** RCS Last Change Date: $Date: 2003/03/20 03:39:47 $
** Original Author: Michael I. Schwartz, mschwart@nyx.net
**
** Note: This is a work in progress. Please coordinate changes with
** the original author.
**
***************************************************************==*/
/****************************************************************
** {Change History}
**
** RCS Change summaries:
**   See printer.log
*****************************************************************/
#include <tcl.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <gd.h>

#ifdef __WIN32__
#include <io.h> /* For access() */
#endif

#ifdef ENABLE_GIF
#include <gdgif.h>
#endif

#include "gdex.h" /* Enable extensions used here */

/* Convert user-oriented angles in degrees to software oriented angles in radians */
#define DEGTORAD(x) ( (x) * (2.0 * 3.14159265) / 360.0 )
#define FIXCARG (char *)

#if TCL_MAJOR_VERSION == 7 && TCL_MINOR_VERSION <= 5
/* In this case, must replace Tcl_Alloc(), Tcl_Realloc(), and Tcl_Free()
** with ckalloc(), ckrealloc(), and ckfree()
*/

#define Tcl_Alloc(x)  ckalloc(x)
#define Tcl_Free(x)   ckfree(x)
#define Tcl_Realloc(x,y)  ckrealloc(x,y)

#endif
   
/*
 * The following functions replace the use of tk.h for certain items that would
 * otherwise require the use of the tk stubs library, which I wish to avoid for
 * this extension.
 */
#ifndef USE_TK_STUBS
   
typedef enum {
    TK_ANCHOR_N, TK_ANCHOR_NE, TK_ANCHOR_E, TK_ANCHOR_SE,
    TK_ANCHOR_S, TK_ANCHOR_SW, TK_ANCHOR_W, TK_ANCHOR_NW,
    TK_ANCHOR_CENTER
} Tk_Anchor;

/*
 * Enumerated type for describing a style of justification:
 */

typedef enum {
    TK_JUSTIFY_LEFT, TK_JUSTIFY_RIGHT, TK_JUSTIFY_CENTER
} Tk_Justify;
   
int Tk_GetJustify(Tcl_Interp *interp, char *string, int *justifyPtr)
{
    if ( strcmp(string,"left") == 0 )
    {
        *justifyPtr = TK_JUSTIFY_LEFT;
        return TCL_OK;
    }
    else if ( strcmp(string,"right") == 0 )
    {
        *justifyPtr = TK_JUSTIFY_RIGHT;
        return TCL_OK;
    }
    else if ( strcmp(string,"center") == 0 )
    {
        *justifyPtr = TK_JUSTIFY_CENTER;
        return TCL_OK;
    }
    Tcl_AppendResult(interp, "Unable to parse justify specification |", string, "|",0);
    return TCL_ERROR;
}

int Tk_GetAnchor(Tcl_Interp *interp, char *string, int *anchorPtr)
{
    if ( strcmp(string,"n") == 0 )
    {
        *anchorPtr = TK_ANCHOR_N;
        return TCL_OK;
    }
    else if ( strcmp(string,"ne") == 0 )
    {
        *anchorPtr = TK_ANCHOR_NE;
        return TCL_OK;
    }
    else if ( strcmp(string,"se") == 0 )
    {
        *anchorPtr = TK_ANCHOR_SE;
        return TCL_OK;
    }
    else if ( strcmp(string,"e") == 0 )
    {
        *anchorPtr = TK_ANCHOR_E;
        return TCL_OK;
    }
    else if ( strcmp(string,"s") == 0 )
    {
        *anchorPtr = TK_ANCHOR_S;
        return TCL_OK;
    }
    else if ( strcmp(string,"center") == 0 )
    {
        *anchorPtr = TK_ANCHOR_CENTER;
        return TCL_OK;
    }
    else if ( strcmp(string,"nw") == 0 )
    {
        *anchorPtr = TK_ANCHOR_NW;
        return TCL_OK;
    }
    else if ( strcmp(string,"w") == 0 )
    {
        *anchorPtr = TK_ANCHOR_W;
        return TCL_OK;
    }
    else if ( strcmp(string,"sw") == 0 )
    {
        *anchorPtr = TK_ANCHOR_SW;
        return TCL_OK;
    }
    Tcl_AppendResult(interp, "Unable to parse anchor specification |", string, "|",0);
    return TCL_ERROR;
}

#else
#include <tk.h>
#endif


#if defined(__WIN32__) || defined(__WIN32S__) || defined(WIN32S) || defined(_MSC_VER) || defined(__BORLANDC__)
  #include <windows.h>
  #   if defined(_MSC_VER)
  #     define EXPORT(a,b) __declspec(dllexport) a b
  #     define DllEntryPoint DllMain
  #   else
  #     if defined(__BORLANDC__)
  #         define EXPORT(a,b) a _export b
  #     else
  #         define EXPORT(a,b) a b
  #     endif
  #   endif
#elif defined(unix)
  #   define EXPORT(a,b) a b
#endif

/* New macros for tcl8.0.3 and later */
#if defined(TCL_STORAGE_CLASS)
#  undef TCL_STORAGE_CLASS
#endif
#define TCL_STORAGE_CLASS DLLEXPORT
#if ! defined(EXTERN)
#  define EXTERN
#endif

/****************************************************************
** Globals
****************************************************************/
/* Try a hash table here.
** The idea is that each image has an initialized structure with a matching name.
** The names are validated in the operations to prevent core dumps.
** In this way many images could be opened at once
** The image name is a required parameter to all GD commands
*/
static void init_image_names(void);
static void delete_image_names(void);
static const char *make_image_name(void);
static int  image_name_valid(const char *name);
static int  image_valid(const char *name);
static int  delete_image_name(const char *name);
static gdImagePtr get_image(const char *name);

EXPORT(int,Gd_Init) (Tcl_Interp *interp);
EXPORT(int,Gd_SafeInit) (Tcl_Interp *interp);

/* Main dispatcher for commands */
static int gd      (ClientData unused, Tcl_Interp *interp, int argc, char **argv);
/* Main dispatcher for subcommands */
static int Gd      (ClientData unused, Tcl_Interp *interp, int argc, char **argv);

/* Individual command handlers */
static int GDOpen      (ClientData unused, Tcl_Interp *interp, int argc, char **argv);
static int GDSave      (ClientData unused, Tcl_Interp *interp, int argc, char **argv);
static int GDClose     (ClientData unused, Tcl_Interp *interp, int argc, char **argv);
static int GDConfig    (ClientData unused, Tcl_Interp *interp, int argc, char **argv);
static int GDFill      (ClientData unused, Tcl_Interp *interp, int argc, char **argv);
static int GDArc       (ClientData unused, Tcl_Interp *interp, int argc, char **argv);
static int GDLine      (ClientData unused, Tcl_Interp *interp, int argc, char **argv);
static int GDOval      (ClientData unused, Tcl_Interp *interp, int argc, char **argv);
static int GDPolygon   (ClientData unused, Tcl_Interp *interp, int argc, char **argv);
static int GDRectangle (ClientData unused, Tcl_Interp *interp, int argc, char **argv);
static int GDText      (ClientData unused, Tcl_Interp *interp, int argc, char **argv);
static int Version     (ClientData unused, Tcl_Interp *interp, int argc, char **argv);

static int GDFontAliases (ClientData unused, Tcl_Interp *interp, int argc, char **argv);

/* Helper functions */
static int color_index(gdImagePtr im, int r, int g, int b);
static int color_select (gdImagePtr im, const char *desc);

static char usage_message[] = "gd [open|close|save|config|fill|"
                              "arc|line|oval|polygon|rectangle|text|"
#if defined(ENABLE_TRUETYPE)
                              "fontalias|"
#endif
                              "version]";
static char msgbuf[1024];

struct gd_fontalias
{
  const char *fontname;
  const char *windows_fontname;
};

static struct gd_fontalias *subst_table;
const char *gdFileFromFont(Tcl_Interp *interp, const char *font, double *ptsize, int debug);

static Tcl_HashTable font_aliases;

static void init_font_aliases(void)
{
  Tcl_InitHashTable(&font_aliases, TCL_STRING_KEYS);
}

static void delete_font_aliases(void)
{
  Tcl_HashEntry *ent;
  Tcl_HashSearch srch;
  char *val;
  for ( ent = Tcl_FirstHashEntry(&font_aliases, &srch); ent != 0;
        ent = Tcl_NextHashEntry(&srch) )
  {
    if ( (val = (char *)Tcl_GetHashValue(ent)) != 0 )
      Tcl_Free(val);
  }
  Tcl_DeleteHashTable(&font_aliases);
}

/* Very simple font matching scheme */
static const char *gdFontFile ( const char *font, int italic, int bold )
{
  /* Add empty choice to bold and italic just in case it isn't supported by the font */
  static const char *boldAliases[] = {
    " Bold", " Demi", " Demibold", " Ultra", "",
  };
  static const char *italicAliases[] = {
    " Italic", " Oblique", " Slant", "",
  };
  static const char *normalAliases[] = {
   " Normal", " Regular", " Book", " Roman", " Plain", " Medium", " Light", 
  };
  Tcl_HashEntry *ent;
    
  Tcl_DString tryfont;
  int i, j;

  /* First, look for an exact match */
  if ( (ent = Tcl_FindHashEntry(&font_aliases, font) ) != 0 )
      return (const char *)Tcl_GetHashValue(ent);

  /* Otherwise, look for common aliases */
  Tcl_DStringInit(&tryfont);
    
  Tcl_DStringAppend(&tryfont, font, -1);

  if ( bold == 0 && italic == 0 )
  {
      for (i=0; i< sizeof(normalAliases) / sizeof (const char *); i++ )
      {
          Tcl_DStringFree(&tryfont);
          Tcl_DStringInit(&tryfont);
          Tcl_DStringAppend(&tryfont, font, -1);
          Tcl_DStringAppend(&tryfont, normalAliases[i], -1 );
          if ( (ent = Tcl_FindHashEntry(&font_aliases, Tcl_DStringValue(&tryfont)) ) != 0 )
          {
              Tcl_DStringFree(&tryfont);
              return (const char *)Tcl_GetHashValue(ent);
          }
      }
  }
    
  if ( bold && italic )
  {
      for (i=0; i< sizeof(boldAliases) / sizeof (const char *); i++ )
      {
          for (j=0; j<sizeof(italicAliases) / sizeof (const char *); j++)
          {
              Tcl_DStringFree(&tryfont);
              Tcl_DStringInit(&tryfont);
              Tcl_DStringAppend(&tryfont, font, -1);
              Tcl_DStringAppend(&tryfont, boldAliases[i], -1 );
              Tcl_DStringAppend(&tryfont, italicAliases[j], -1 );
              if ( (ent = Tcl_FindHashEntry(&font_aliases, Tcl_DStringValue(&tryfont)) ) != 0 )
              {
                  Tcl_DStringFree(&tryfont);
                  return (const char *)Tcl_GetHashValue(ent);
              }
          }
      }
  }
  else if (bold)
  {
      for (i=0; i< sizeof(boldAliases) / sizeof (const char *); i++ )
      {
          Tcl_DStringFree(&tryfont);
          Tcl_DStringInit(&tryfont);
          Tcl_DStringAppend(&tryfont, font, -1);
          Tcl_DStringAppend(&tryfont, boldAliases[i], -1 );
          if ( (ent = Tcl_FindHashEntry(&font_aliases, Tcl_DStringValue(&tryfont)) ) != 0 )
          {
              Tcl_DStringFree(&tryfont);
              return (const char *)Tcl_GetHashValue(ent);
          }
      }
  }
  else if (italic)
  {
      for (i=0; i< sizeof(italicAliases) / sizeof (const char *); i++ )
      {
          Tcl_DStringFree(&tryfont);
          Tcl_DStringInit(&tryfont);
          Tcl_DStringAppend(&tryfont, font, -1);
          Tcl_DStringAppend(&tryfont, italicAliases[i], -1 );
          if ( (ent = Tcl_FindHashEntry(&font_aliases, Tcl_DStringValue(&tryfont)) ) != 0 )
          {
              Tcl_DStringFree(&tryfont);
              return (const char *)Tcl_GetHashValue(ent);
          }
      }
  }
    
  return 0;
}

static int gdFontSetArray ( Tcl_Interp *interp, const char *arrayname )
{
    int i=0;
    /* I guess we'll just do "array names" to get the list of names */
    if ( Tcl_VarEval(interp, "lsort [ array names ", arrayname, " ]", (char *)0 ) == TCL_OK )
    {
        char **argv;
        int  argc;
        char *string;
        if ( (string = Tcl_GetStringResult(interp)) != 0 )
        {
            int i;
            Tcl_SplitList(interp, string, &argc, &argv);
            for (i=0; i<argc; i++)
            {
                char *key = argv[i];
                char *value;
                if ( (value = Tcl_GetVar2(interp, (char *)arrayname, key, 0)) != 0 )
                {
                    char *val = Tcl_Alloc(strlen(value)+1); /* Should this be Tcl_DuplicateObj? */
                    if (val)
                    {
                        int status;
                        Tcl_HashEntry *data;
                        strcpy(val, value);
                        Tcl_CreateHashEntry(&font_aliases, key, &status);
                        if ( (data = Tcl_FindHashEntry(&font_aliases, key) ) != 0 )
                            Tcl_SetHashValue(data, val);
                    }
                }
            }
        }
        
    }
    
    return i;
}

static int gdFontGetArray ( Tcl_Interp *interp, const char *arrayname )
{
  int i=0;
    
  Tcl_HashEntry *ent;
  Tcl_HashSearch srch;
  char *val;
  for ( ent = Tcl_FirstHashEntry(&font_aliases, &srch); ent != 0;
        ent = Tcl_NextHashEntry(&srch) )
  {
    i++;
    if ( (val = (char *)Tcl_GetHashValue(ent)) != 0 )
      Tcl_SetVar2(interp, (char *)arrayname, Tcl_GetHashKey(&font_aliases, ent), val, 0);
  }
    
  return i;
}

static int strtrim(char *str)
{
    int i;
    int j = 0;
    int seenflag = 0;
    
    for (i=0; str[i] != '\0'; i++)
    {
        if ( seenflag == 0 )
        {
            if ( str[i] != ' ' && str[i] != '\t' && str[i] != '\r' && str[i] != '\n' && str[i] != '\f' )
            {
                seenflag = 1;
                str[j++] = str[i];
            }
        }
        else
            str[j++] = str[i];
    }
    str[j] = '\0';
    
    for (i=j-1; i>0; i--)
    {
        if ( str[i] == ' ' || str[i] == '\t' || str[i] == '\r' || str[i] == '\n' || str[i] == '\f' )
            str[i] = '\0';
        else
            break;
    }
    return i+1;
}

static int gdFirstWhite(const char *str)
{
    const char *cp = strpbrk(str," \t\f\r\n");
    if ( cp )
        return cp - str;
    return -1;
}

static int gdFirstBreak(const char *str)
{
    const char *cp = strpbrk(str,"\f\r\n");
    if ( cp )
        return cp - str;
    return -1;
}

static int gdFirstSpace(const char *str)
{
    const char *cp = strpbrk(str," \t");
    if ( cp )
        return cp - str;
    return -1;
}

static int gdLastSpaceBefore(const char *str, int ind)
{
    int retval = -1;
    char *cp;
    
    while ( (cp = strpbrk(str+retval+1, " \t") ) != NULL )
    {
        if ( (cp - str) < ind )
            retval = cp - str;
        else
            break;
    }
    return retval;
}

static const char *gdSkipLeadingWhite(const char *str)
{
    int len = strspn(str," \t\r\n\f");
    return str + len;
}

/****************************************************************
** This is the top-level routine for the GD command
****************************************************************/
static int gd  (ClientData unused, Tcl_Interp *interp, int argc, char **argv)
{

  if ( argc > 1 && strcmp(*argv, "gd") == 0 )
  {
    argc--;
    argv++;
    return Gd(unused, interp, argc, argv);
  }

  Tcl_SetResult (interp, usage_message, TCL_STATIC);
  return TCL_ERROR;
}

struct gdi_command
{
  char *command_string;
  int (*command) (ClientData, Tcl_Interp *, int, char **);
} gdi_commands[] =
{
  { "arc",        GDArc },
  { "close",      GDClose },
  { "config",     GDConfig },
  { "fill",       GDFill },
#if defined(ENABLE_TRUETYPE)
  { "fontalias",  GDFontAliases },
#endif
  { "line",       GDLine },
  { "open",       GDOpen },
  { "oval",       GDOval },
  { "polygon",    GDPolygon },
  { "rectangle",  GDRectangle },
  { "save",       GDSave },
  { "text",       GDText },
  { "version",    Version },
  
};

/****************************************************************
** This is the GD subcommand dispatcher
****************************************************************/
static int Gd      (ClientData unused, Tcl_Interp *interp, int argc, char **argv)
{
  int i;

  for (i=0; i<sizeof(gdi_commands) / sizeof(struct gdi_command); i++)
    if ( strcmp (*argv, gdi_commands[i].command_string) == 0 )
      return (*gdi_commands[i].command)(unused, interp, argc-1, argv+1);

  Tcl_SetResult (interp, usage_message, TCL_STATIC);
  return TCL_ERROR;
}

/****************************************************************
** GDFill - Tcl procedure to fill an area with a given color
**          It will fill to the outline color if given,
**          otherwise it will fill only the color at the given
**          point
****************************************************************/
static int GDFill      (ClientData unused, Tcl_Interp *interp, int argc, char **argv)
{
  int x, y;
  int i;
  gdImagePtr im;
  int color;
  int do_outline = 0;
  int do_color = 0;
  int fill;
  static char usage[] = "gd fill image x y [-outline color] -fill color";

  if (argc < 5)
  {
    Tcl_SetResult(interp, usage, TCL_STATIC);
    return TCL_ERROR;
  }

  if ( (im = get_image(argv[0])) == 0 )
  {
    Tcl_AppendResult(interp, "Image error: ", argv[0], " is not a valid GD image", NULL);
    return TCL_ERROR;
  }

  x = atoi(argv[1]);
  y = atoi(argv[2]);
  argc -= 3;
  argv += 3;

  for (i=0; i<argc;i++)
  {
    if ( strcmp(argv[i], "-outline") == 0 )
    {
      if ( ++i < argc)
      {
        do_outline = 1;
        color = color_select(im, argv[i]);
      }
    }
    else if ( strcmp(argv[i], "-fill") == 0 )
    {
      if ( ++i < argc)
      {
        do_color = 1;
        /****************
        ** Sigh. Exclude the non-working cases
        ****************/
        if ( argv[i][0] == '@' )
        {
          Tcl_SetResult(interp, 
                        "The GD library does not properly handle fills "
                        "with brushes at this time", TCL_STATIC);
          return TCL_ERROR;
        }
        else if ( argv[i][0] == '%' )
        {
          Tcl_SetResult(interp, 
                        "The GD library does not properly handle fills "
                        "with tiles at this time", TCL_STATIC);
          return TCL_ERROR;
        }
        fill = color_select(im, argv[i]);
      }
    }
  }

  if (do_color == 0)
  {
    Tcl_SetResult(interp, usage, TCL_STATIC);
    return TCL_ERROR;
  }
  
  if (do_outline)
    gdImageFillToBorder(im, x, y, color, fill);
  else
    gdImageFill(im, x, y, fill);

  return TCL_OK;
}

/****************************************************************
** Set up a family of arc-drawing primitives
****************************************************************/
typedef int (*DrawFunc) (const char *, int, int, int, int, int, int, int, int);
static int lArc(const char *name, int xul, int yul, int xlr, int ylr,
                                 int sangle,  int extent, int color, int thick);
static int lPie(const char *name, int xul, int yul, int xlr, int ylr,
                                 int sangle,  int extent, int color, int thick);
static int lChord(const char *name, int xul, int yul, int xlr, int ylr,
                                 int sangle,  int extent, int color, int thick);

/****************************************************************
** GDArc - Draw arcs in pieslice, chord, or unfilled style.
**         If a fill is requested, change "unfilled" to "chord"
****************************************************************/
static int GDArc      (ClientData unused, Tcl_Interp *interp, int argc, char **argv)
{
  int x1, y1, x2, y2;
  char *image;
  double extent = 0.0 , start = 0.0 ;
  int color = 0;
  int fill;
  int dofill=0;
  int width = 1;
  DrawFunc drawfunc;
  
  static char usage_message[] = "gd arc image x1 y1 x2 y2 "
                                "-extent degrees "
                                "-fill color -outline color "
                                "-outlinestipple bitmap "
                                "-start degrees -stipple bitmap "
                                "-style [pieslice|chord|arc] -width linewid";

  gdImagePtr im;
  
  drawfunc = lPie;
  
  /* Verrrrrry simple for now... */
  if (argc >= 5)
  {
    image = argv[0];
    /* One check for validity... */
    if ( (im = get_image(image)) == 0 )
    {
      Tcl_AppendResult(interp, "Image error: ", image, " is not a valid GD image", NULL);
      return TCL_ERROR;
    }
    
    x1 = atoi(argv[1]);
    y1 = atoi(argv[2]);
    x2 = atoi(argv[3]);
    y2 = atoi(argv[4]);
    if ( x1 > x2 ) { int x3 = x1; x1 = x2; x2 = x3; }
    if ( y1 > y2 ) { int y3 = y1; y1 = y2; y2 = y3; }
    argc -= 5;
    argv += 5;
    while ( argc >= 2 )
    {
      if ( strcmp (argv[0], "-extent") == 0 )
        extent = atof(argv[1]);
      else if ( strcmp (argv[0], "-start") == 0 )
        start  = atof(argv[1]);
      else if ( strcmp (argv[0], "-style") == 0 )
      {
        if ( strcmp (argv[1], "pieslice") == 0 )
          drawfunc = lPie;
        else if ( strcmp(argv[1], "arc") == 0 )
          drawfunc = lArc;
        else if ( strcmp(argv[1], "chord") == 0 )
          drawfunc = lChord;
      }
      else if ( strcmp(argv[0], "-outline") == 0 ) /* line color */
        color = color_select(im, argv[1]);
      else if ( strcmp(argv[0], "-width") == 0 ) /* line width */
      {
        if ( (width = atoi(argv[1])) == 0 )
          width = 1;
      }
      else if ( strcmp(argv[0], "-fill") == 0 ) /* fill color */
      {
        dofill = 1;
        fill = color_select(im, argv[1]);
      }
      else if ( strcmp(argv[0], "-outlinestipple" ) == 0 ) /* unsupported */
        ;
      else if ( strcmp(argv[0], "-stipple" ) == 0 ) /* unsupported */
        ;

      argc -= 2;
      argv += 2;
    }

    if (dofill && drawfunc == lArc)
      drawfunc = lChord;

    (*drawfunc)(image, x1, y1, x2, y2, (int)start, (int)extent, color, width);

    /* 
    ** Yes, I know this next item is really a kludge since internal
    ** shapes with borders of the same color won't get filled, but
    ** this is the best available without changing libgd
    */
    if ( dofill )
    {
      int cx, cy, w, h;
      int angle = (int)(start + ( extent / 2 ));
      int xr0, yr0;

      cx = (x1 + x2) / 2;
      cy = (y1 + y2) / 2;
      w  = (x2 - x1);
      h  = (y2 - y1);
      xr0 = cx;
      yr0 = cy;
      xr0 += (int)(w * cos( DEGTORAD(angle) ) / 2.0) ;
      yr0 += (int)(h * sin( DEGTORAD(angle) ) / 2.0) ;

      if ( xr0 - cx > 0 ) xr0-= width ; else if (xr0 - cx < 0 ) xr0+=width;
      if ( yr0 - cy > 0 ) yr0-= width ; else if (xr0 - cy < 0 ) yr0+=width;
      
      gdImageFillToBorder(im, xr0, yr0, color, fill);
#if 0
      /* For now, just to see where the fill point is... */
      gdImageFilledRectangle(im, xr0-5, yr0-5, xr0+5, yr0+5, color);
#endif
    }

    return TCL_OK;
  }

  Tcl_SetResult(interp, usage_message, TCL_STATIC);
  return TCL_ERROR;
}

/****************************************************************
** GDLine - 
****************************************************************/
static int GDLine     (ClientData unused, Tcl_Interp *interp, int argc, char **argv)
{
  static char usage_message[] = "gd line image x1 y1 ... xn yn "
                                "-arrow [first|last|both|none] -arrowshape {d1 d2 d3} "
                                "-capstyle [butt|projecting|round] -fill color "
                                "-joinstyle [bevel|miter|round] -smooth [true|false] "
                                "-splinesteps number -stipple bitmap -width linewid";
  char *strend;
  gdPoint *polypoints;
  int npoly;
  int x, y;
  gdImagePtr im;
  int color = 0;
  int width = 1;
  int doarrow = 0; /* 0=none; 1=end; 2=start; 3=both */
  int arrowshape[3];
  
  arrowshape[0] = 8;
  arrowshape[1] = 10;
  arrowshape[2] = 3;

  /* Verrrrrry simple for now... */
  if (argc >= 5)
  {
    im = get_image(argv[0]);
    if ( im == 0 )
    {
      Tcl_AppendResult(interp, "Image error: ", argv[0], " is not a valid image name", NULL);
      return TCL_ERROR;
    }

    if ( (polypoints = (gdPoint *)Tcl_Alloc(argc * sizeof(gdPoint))) == 0 )
    {
      Tcl_SetResult(interp, "Out of memory in GdLine", TCL_STATIC);
      return TCL_ERROR;
    }
    polypoints[0].x = atoi(argv[1]);
    polypoints[0].y = atoi(argv[2]);
    polypoints[1].x = atoi(argv[3]);
    polypoints[1].y = atoi(argv[4]);
    argc -= 5;
    argv += 5;
    npoly = 2;
    
    while ( argc >= 2 )
    {
      /* Check for a number  */
      x = strtoul(argv[0], &strend, 0);
      if ( strend > argv[0] )
      {
        /* One number... */
        y = strtoul (argv[1], &strend, 0);
        if ( strend > argv[1] )
        {
          /* TWO numbers! */
          polypoints[npoly].x = x;
          polypoints[npoly].y = y;
          npoly++;
          argc-=2;
          argv+=2;
        }
        else
        {
          /* Only one number... Assume a usage error */
          Tcl_Free((void *)polypoints);
          Tcl_SetResult(interp, usage_message, TCL_STATIC);
          return TCL_ERROR;
        }
      }
      else
      {
        /* Check for arguments
        ** Most of the arguments affect the "Pen"
        */
        if ( strcmp(argv[0], "-fill") == 0 ) /* Fill color */
        {
          argv++;
          argc--;
          color = color_select(im, argv[0]);
        }
        else if ( strcmp(argv[0], "-width") == 0 )
        {
          argc--;
          argv++;
          width = atoi(argv[0]); 
        }
        else if ( strcmp(argv[0], "-arrow" ) == 0 ) /* unsupported */
        {
          if ( strcmp(argv[1], "none") == 0 )
            doarrow = 0;
          else if ( strcmp(argv[1], "both") == 0 )
            doarrow = 3;
          else if ( strcmp(argv[1], "first") == 0 )
            doarrow = 2;
          else if ( strcmp(argv[1], "last") == 0 )
            doarrow = 1;

          argc--;
          argv++;
        }
        else if ( strcmp(argv[0], "-arrowshape") == 0 ) /* unsupported */
        {
          /* List of 3 numbers--set arrowshape array */
          int a1, a2, a3;

          if ( sscanf(argv[1], "%d%d%d", &a1, &a2, &a3) == 3 )
          {
            if (a1 > 0 && a2 > 0 && a3 > 0 )
            {
              arrowshape[0] = a1;
              arrowshape[1] = a2;
              arrowshape[2] = a3;
            }
            /* Else the numbers are bad */
          }
          /* Else the argument was bad */
          
          argc--;
          argv++;
        }
        else if ( strcmp( argv[0], "-capstyle") == 0 ) /* unsupported */
        {
          argc--;
          argv++;
        }
        else if ( strcmp( argv[0], "-joinstyle") == 0 ) /* unsupported */
        {
          argc--;
          argv++;
        }
        else if ( strcmp( argv[0], "-smooth") == 0 ) /* unsupported */
        {
          argc--;
          argv++;
        }
        else if ( strcmp( argv[0], "-splinesteps") == 0 ) /* unsupported */
        {
          argc--;
          argv++;
        }
        else if ( strcmp( argv[0], "-stipple") == 0 ) /* unsupported */
        {
          argc--;
          argv++;
        }
        argc--;
        argv++;
      }
    }

    if (npoly > 2)
      gdImageOpenPolygonEx(im, polypoints, npoly, color, width);
    else
      gdImageLineEx(im, polypoints[0].x, polypoints[0].y, polypoints[1].x, polypoints[1].y, color, width);

    /* Now the arrowheads if any */
    if ( doarrow & 1 )
    {
      gdPoint ahead [6];
      /* Arrowhead at end = polypoints[npoly-1].x, polypoints[npoly-1].y */
      double dx, dy, length;
      double backup, sinTheta, cosTheta;
      double vertX, vertY, temp;
      double fracHeight;

      fracHeight = 2.0 / arrowshape[2];
      backup = fracHeight*arrowshape[1] + arrowshape[0]*(1.0 - fracHeight)/2.0;
      
      ahead[0].x = ahead[5].x = polypoints[npoly-1].x;
      ahead[0].y = ahead[5].y = polypoints[npoly-1].y;
      dx = ahead[0].x - polypoints[npoly-2].x;
      dy = ahead[0].y - polypoints[npoly-2].y;
      if ( (length = hypot(dx, dy)) == 0 )
        sinTheta = cosTheta = 0.0;
      else
      {
        sinTheta = dy / length;
        cosTheta = dx / length;
      }
      vertX = ahead[0].x - arrowshape[0]*cosTheta;
      vertY = ahead[0].y - arrowshape[0]*sinTheta;
      temp = arrowshape[2]*sinTheta;
      ahead[1].x = (int)(ahead[0].x - arrowshape[1]*cosTheta + temp);
      ahead[4].x = (int)(ahead[1].x - 2 * temp);
      temp = arrowshape[2]*cosTheta;
      ahead[1].y = (int)(ahead[0].y - arrowshape[1]*sinTheta - temp);
      ahead[4].y = (int)(ahead[1].y + 2 * temp);
      ahead[2].x = (int)(ahead[1].x*fracHeight + vertX*(1.0-fracHeight));
      ahead[2].y = (int)(ahead[1].y*fracHeight + vertY*(1.0-fracHeight));
      ahead[3].x = (int)(ahead[4].x*fracHeight + vertX*(1.0-fracHeight));
      ahead[3].y = (int)(ahead[4].y*fracHeight + vertY*(1.0-fracHeight));

      gdImageFilledPolygon(im, ahead, 6, color);   /* Always fill */
#if 0
      gdImagePolygonEx(im, ahead, 6, color, width);  /* Always draw border (?) */
#endif
    }
    if ( doarrow & 2 )
    {
      /* Arrowhead at end = polypoints[0].x, polypoints[0].y */
      gdPoint ahead[6];
      double dx, dy, length;
      double backup, sinTheta, cosTheta;
      double vertX, vertY, temp;
      double fracHeight;

      fracHeight = 2.0 / arrowshape[2];
      backup = fracHeight*arrowshape[1] + arrowshape[0]*(1.0 - fracHeight)/2.0;
      
      ahead[0].x = ahead[5].x = polypoints[0].x;
      ahead[0].y = ahead[5].y = polypoints[0].y;
      dx = ahead[0].x - polypoints[1].x;
      dy = ahead[0].y - polypoints[1].y;
      if ( (length = hypot(dx, dy)) == 0 )
        sinTheta = cosTheta = 0.0;
      else
      {
        sinTheta = dy / length;
        cosTheta = dx / length;
      }
      vertX = ahead[0].x - arrowshape[0]*cosTheta;
      vertY = ahead[0].y - arrowshape[0]*sinTheta;
      temp = arrowshape[2]*sinTheta;
      ahead[1].x = (int)(ahead[0].x - arrowshape[1]*cosTheta + temp);
      ahead[4].x = (int)(ahead[1].x - 2 * temp);
      temp = arrowshape[2]*cosTheta;
      ahead[1].y = (int)(ahead[0].y - arrowshape[1]*sinTheta - temp);
      ahead[4].y = (int)(ahead[1].y + 2 * temp);
      ahead[2].x = (int)(ahead[1].x*fracHeight + vertX*(1.0-fracHeight));
      ahead[2].y = (int)(ahead[1].y*fracHeight + vertY*(1.0-fracHeight));
      ahead[3].x = (int)(ahead[4].x*fracHeight + vertX*(1.0-fracHeight));
      ahead[3].y = (int)(ahead[4].y*fracHeight + vertY*(1.0-fracHeight));

      gdImageFilledPolygon(im, ahead, 6, color);   /* Always fill */
#if 0
      gdImagePolygonEx(im, ahead, 6, color, width);  /* Always draw border (?) */
#endif
    }
    
    Tcl_Free((void *)polypoints);
    
    return TCL_OK;
  }

  Tcl_SetResult(interp, usage_message, TCL_STATIC);
  return TCL_ERROR;
}

/******************************************************************************
** GDOval - Ellipses
******************************************************************************/
static int GDOval     (ClientData unused, Tcl_Interp *interp, int argc, char **argv)
{
  static char usage_message[] = "gd oval image x1 y1 x2 y2 -fill color -outline color "
                                "-stipple bitmap -width linewid";
  int x1, y1, x2, y2;
  int color = 0;
  gdImagePtr im;
  int dofill = 0;
  int fill;
  int width = 1;
  
  /* Verrrrrry simple for now... */
  if (argc >= 5)
  {
    if ( (im = get_image(argv[0])) == 0 )
    {
      Tcl_AppendResult(interp, "Image error: ", argv[0], " is not a valid GD name", NULL);
      return TCL_ERROR;
    }

    x1 = atoi(argv[1]);
    y1 = atoi(argv[2]);
    x2 = atoi(argv[3]);
    y2 = atoi(argv[4]);
    if ( x1 > x2 ) { int x3 = x1; x1 = x2; x2 = x3; }
    if ( y1 > y2 ) { int y3 = y1; y1 = y2; y2 = y3; }
    argc -= 5;
    argv += 5;
    
    /* Now handle any other arguments that occur */
    while ( argc > 0 )
    {
      if ( strcmp(argv[0], "-outline") == 0 ) /* Outline color */
      {
        argv++;
        argc--;
        if ( argc > 0 )
          color = color_select(im, argv[0]);
      }
      else if ( strcmp(argv[0], "-fill") == 0 ) /* fill color */
      {
        argv++;
        argc--;
        if ( argc > 0 )
        {
          fill = color_select(im, argv[0]);
          dofill = 1;
        }
      }
      else if ( strcmp(argv[0], "-width") == 0 ) /* line width */
      {
        argv++;
        argc--;
        if ( argc > 0 )
          width = atoi(argv[0]);
      }
      else if ( strcmp(argv[0], "-stipple") == 0 ) /* unsupported */
      {
        argv++;
        argc--;
      }
      argc--;
      argv++;
    }
    gdImageEllipseEx(im, (x1+x2)/2, (y1+y2)/2, x2-x1, y2-y1, color, width);
    /* Yes, I know this next item is really a kludge since internal
    ** shapes with borders of the same color won't get filled, but
    ** this is the best available without changing libgd
    */
    if (dofill)
      gdImageFillToBorder(im, (x1+x2)/2, (y1+y2)/2, color, fill);
    return TCL_OK;
  }

  Tcl_SetResult(interp, usage_message, TCL_STATIC);
  return TCL_ERROR;
}

/******************************************************************************
** GDPolygon - closed line path
******************************************************************************/
static int GDPolygon  (ClientData unused, Tcl_Interp *interp, int argc, char **argv)
{
  static char usage_message[] = "gd polygon image x1 y1 ... xn yn "
                                "-fill color -outline color -smooth [true|false] "
                                "-splinesteps number -stipple bitmap -width linewid";

  char *strend;
  gdPoint *polypoints;
  int npoly;
  int x, y;
  int color = 0;
  gdImagePtr im;
  int do_fill = 0;
  int fill;
  int width = 1;
  
  /* Verrrrrry simple for now... */
  if (argc >= 5)
  {
    if ( (im = get_image(argv[0])) == 0 )
    {
      Tcl_AppendResult(interp, "Image error: ", argv[0], " is not a valid GD image", NULL);
      return TCL_ERROR;
    }

    if ( (polypoints = (gdPoint *)Tcl_Alloc(argc * sizeof(gdPoint))) == 0 )
    {
      Tcl_SetResult(interp, "Out of memory in GDPolygon", TCL_STATIC);
      return TCL_ERROR;
    }
    polypoints[0].x = atoi(argv[1]);
    polypoints[0].y = atoi(argv[2]);
    polypoints[1].x = atoi(argv[3]);
    polypoints[1].y = atoi(argv[4]);
    argc -= 5;
    argv += 5;
    npoly = 2;
    
    while ( argc >= 2 )
    {
      /* Check for a number  */
      x = strtoul(argv[0], &strend, 0);
      if ( strend > argv[0] )
      {
        /* One number... */
        y = strtoul (argv[1], &strend, 0);
        if ( strend > argv[1] )
        {
          /* TWO numbers! */
          polypoints[npoly].x = x;
          polypoints[npoly].y = y;
          npoly++;
          argc-=2;
          argv+=2;
        }
        else
        {
          /* Only one number... Assume a usage error */
          Tcl_Free((void *)polypoints);
          Tcl_SetResult(interp, usage_message, TCL_STATIC);
          return TCL_ERROR;
        }
      }
      else
      {
        /* Check for arguments
        ** Most of the arguments affect the "Pen" and "Brush"
        */
        if ( strcmp(argv[0], "-outline") == 0 ) /* Fill color */
        {
          argc--;
          argv++;
          color = color_select(im, argv[0]);
        }
        else if ( strcmp (argv[0], "-fill") == 0 )
        {
          argc--;
          argv++;
          if ( argc > 0 )
          {
            do_fill = 1;
            fill = color_select(im, argv[0]);
          }
        }
        else if ( strcmp (argv[0], "-width") == 0 )
        {
          argc--;
          argv++;
          if ( argc > 0 )
            width = atoi(argv[0]);
        }
        else if ( strcmp (argv[0], "-smooth") == 0 ) /* unsupported */
        {
          argc--;
          argv++;
        }
        else if ( strcmp (argv[0], "-splinesteps") == 0 ) /* unsupported */
        {
          argc--;
          argv++;
        }
        else if ( strcmp (argv[0], "-stipple") == 0 ) /* unsupported */
        {
          argc--;
          argv++;
        }
        argc--;
        argv++;
      }
    }

    if (do_fill)
      gdImageFilledPolygon(im, polypoints, npoly, fill);
    gdImagePolygonEx(im, polypoints, npoly, color, width);
    Tcl_Free((void *)polypoints);
    
    return TCL_OK;
  }

  Tcl_SetResult(interp, usage_message, TCL_STATIC);
  return TCL_ERROR;
}

/******************************************************************************
** GDRectangle - an orthorectified rectangle
******************************************************************************/
static int GDRectangle(ClientData unused, Tcl_Interp *interp, int argc, char **argv)
{
  static char usage_message[] = "gd rectangle image x1 y1 x2 y2 "
                                "-fill color -outline color "
                                "-stipple bitmap -width linewid";

  int x1, y1, x2, y2;
  int color = 0;
  gdImagePtr im;
  int fill = 0;
  int do_fill = 0;
  int width = 1;
  
  /* Verrrrrry simple for now... */
  if (argc >= 5)
  {
    if ( (im = get_image(argv[0])) == 0 )
    {
      Tcl_AppendResult(interp, "Image error: ", argv[0], " is not a valid GD image", NULL);
      return TCL_ERROR;
    }

    x1 = atoi(argv[1]);
    y1 = atoi(argv[2]);
    x2 = atoi(argv[3]);
    y2 = atoi(argv[4]);
    if ( x1 > x2 ) { int x3 = x1; x1 = x2; x2 = x3; }
    if ( y1 > y2 ) { int y3 = y1; y1 = y2; y2 = y3; }
    argc -= 5;
    argv += 5;
    
    /* Now handle any other arguments that occur */
    while (argc > 0 )
    {
      if ( strcmp(argv[0], "-outline") == 0 ) /* Fill color */
      {
        argc--;
        argv++;
        if ( argc > 0 )
          color = color_select(im, argv[0]);
      }
      else if ( strcmp(argv[0], "-fill") == 0 )
      {
        argc--;
        argv++;
        if ( argc > 0 )
        {
          do_fill = 1;
          fill = color_select(im, argv[0]);
        }
      }
      else if ( strcmp(argv[0], "-width") == 0 )
      {
        argc--;
        argv++;
        if ( argc > 0 )
          width = atoi(argv[0]);
      }
      else if ( strcmp(argv[0], "-stipple") == 0 ) /* unsupported */
      {
        argc--;
        argv++;
      }
      argc--;
      argv++;
    }

    /* Fill first */
    if (do_fill)
      gdImageFilledRectangle (im, x1, y1, x2, y2, fill);
    /* Outline second */
    gdImageRectangleEx (im, x1, y1, x2, y2, color, width);

    return TCL_OK;
  }

  Tcl_SetResult(interp, usage_message, TCL_STATIC);
  return TCL_ERROR;
}

/* Helpful constants */
/* Bits: 11 */
/* 1=Alignment */

#define GDX_ALIGNMENT(x)  ((x)&0x3)
#define GDX_LEFT    0x0000
#define GDX_RIGHT   0x0001
#define GDX_CENTER  0x0002
#define GDX_JUSTIFY 0x0003

/* Helper functions */

static int gdFitWidthFixed(const char *init, int width, gdFontPtr font);
static int gdFitWidthTTF(Tcl_Interp *interp, const char *init, int *width, const char *fontfile, double fontsize);
static int gdSingleLine(const char *init, char *line, int len);

static int gdTextLineFixed(Tcl_Interp *interp,
                           gdImagePtr im, gdFontPtr font, 
                           int x0, int y0, 
                           int color, int debug,
                           const char *string, int len, double angle,
                           int *wid, int *hgt);
static int gdTextLineTTF(Tcl_Interp *interp,
                           gdImagePtr im, const char *fontfile, double fontsize, 
                           int x0, int y0, 
                           int color, int debug,
                           const char *string, int len, double angle,
                           int *wid, int *hgt);
static int gdTextBlockFixed(Tcl_Interp *interp,
                           gdImagePtr im, gdFontPtr font, 
                           int x0, int y0, 
                           int color, int debug,
                           const char *string, int anchorflags, int justifyflags,
                           double angle,
                           int *wid, int *hgt);
static int gdTextBlockTTF(Tcl_Interp *interp,
                           gdImagePtr im, const char *fontfile, double fontsize, 
                           int x0, int y0, 
                           int color, int debug,
                           const char *string, 
                           int anchorflags, int justifyflags,
                           double angle,
                           int *wid, int *hgt);

/****************************************************************
** This is a completely new implmentation of GDText as of 0.3.0.1.
** While currently anchor and justify are not respected, this should
** be possible to do, and do consistently between font types.
****************************************************************/                          
static int GDText     (ClientData unused, Tcl_Interp *interp, int argc, char **argv)
{
  static char usage_message[] = "gd text image x y -anchor [center|n|e|s|w] "
                                "-fill color -font fontname "
                                "-justify [left|right|center] "
                                "-stipple bitmap -text string -width linelen"
                                "-orientation horizontal|vertical "
                                "-rotate degrees -single "
                                "[-render|-norender] "
                                "[-antialias|-noantialias] "
                                "-debug";

  gdImagePtr im;
  int x, y;
  char *string = 0;
  int orient = 0; /* 0 - horizontal; 1 - vertical */
  double angle = 0.0;
  char *fontname=0;
  const char *fontfile=0;
  double fontsize = 10.0;
  extern gdFontPtr gdFontLarge, gdFontSmall, gdFontGiant, gdFontTiny, gdFontMediumBold;
  gdFontPtr font = gdFontMediumBold;
  int color = 0;
  int doantialias = 1; /* For truetype fonts, positive fontsize */
  int do_truetype = 0;
  int do_render = 1;
  int do_debug = 0;
  int width = 0;
  int singleline = 0;
  int wid;
  int hgt;
  int anchorstyle =  TK_ANCHOR_CENTER; /* Appears to be the default */
  int justifystyle = TK_JUSTIFY_LEFT;  /* 0=left; 2=center; 1=right */
              
  if ( argc >= 4 )
  {
    /* Parse the command */
    if ( (im = get_image(argv[0])) == 0 )
    {
      Tcl_AppendResult(interp, "Image error: ", argv[0], " is not a valid GD image", NULL);
      return TCL_ERROR;
    }
    x = atoi(argv[1]);
    y = atoi(argv[2]);
    argc -= 3;
    argv += 3;

    while ( argc > 0 )
    {
      if ( strcmp(argv[0], "-anchor") == 0 )
      {
        int val;
          
        argc--;
        argv++;
        /* Note: I'll use Tk_GetAnchor() here so long as it doesn't add tk dependency
         * to everything. */
        if ( argv[0] == 0 || (val = Tk_GetAnchor(interp, argv[0], &anchorstyle)) != TCL_OK )
        {
            /* The anchor specification was not understood */
            /* Possibilities are: TK_ANCHOR_CENTER, N, NE, NW, E, W, S, SE, SW */
            return TCL_ERROR;
        }
      }
      else if ( strcmp(argv[0], "-debug") == 0 )
        do_debug = 1;
      else if ( strcmp(argv[0], "-justify") == 0 )
      {
        int val;
          
        argc--;
        argv++;
        /* Note: I'll use Tk_GetAnchor() here so long as it doesn't add tk dependency
         * to everything. */
        if ( argv[0] == 0 || (val = Tk_GetJustify(interp, argv[0], &justifystyle)) != TCL_OK )
        {
            /* The justify specification was not understood */
            /* Possibilities are: TK_JUSTIFY_LEFT, TK_JUSTIFY_CENTER, TK_JUSTIFY_LEFT */
            return TCL_ERROR;
        }
      }
      else if ( strcmp(argv[0], "-text") == 0 )
      {
        /* This string will have to be cast into the right character set... */
        argc--;
        argv++;
        if (argc > 0 )
          string = argv[0];
      }
      else if ( strcmp(argv[0], "-width") == 0 )
      {
        argc--;
        argv++;
        if ( argc > 0 )
          width = atoi(argv[0]);
      }
      else if ( strcmp(argv[0], "-orientation") == 0 )
      {
        argc--;
        argv++;
        if (argc > 0 )
          switch (argv[0][0])
          {
            case 'h': case 'H':
              orient = 0;
              angle  = 0.0;
              break;
            case 'v': case 'V':
              orient = 1;
              angle = 90.0;
              break;
          }
      }
      else if ( strcmp(argv[0], "-rotate") == 0 )
      {
        argv++;
        argc--;
        if (argc > 0 )
        {
          char *strend;
          double ang;
          ang = strtod (argv[0], &strend);
          if ( strend != 0 && strend != argv[0] )
          {
            angle = ang;
            if ( angle == 0.0 )
              orient = 0;
            if ( angle == 90.0 )
              orient = 1;
          }
        }
      }
      else if ( strcmp(argv[0], "-fill") == 0 )
      {
        argv++;
        argc--;
        if (argc > 0)
          color = color_select(im, argv[0]);
      }
      else if ( strcmp(argv[0], "-font") == 0 )
      {
        argc--;
        argv++;
        fontname = argv[0];
        /* For the specifically names fonts, ignore size, etc. */
        if ( strncmp( fontname, "tiny", 4) == 0 )
          font = gdFontTiny;
        else if ( strncmp(fontname, "small", 5) == 0 )
          font = gdFontSmall;
        else if ( strncmp(fontname, "medium", 6) == 0 )
          font = gdFontMediumBold;
        else if ( strncmp(fontname, "large", 5) == 0 )
          font = gdFontLarge;
        else if ( strncmp(fontname, "giant", 5) == 0 )
          font = gdFontGiant;
        else      
        {
          char *err;
          int brect[8];
          const char *fn;
          /* Try a TrueType font */
#if defined(ENABLE_TRUETYPE)
          fn = gdFileFromFont(interp, fontname, &fontsize, do_debug);
          if ( do_debug)
            Tcl_AppendResult(interp, "{DEBUG: Truetype font: ", argv[0], " gives file ", fn, "}\n", 0);
            
          /* See if the font exists */
#if defined(OLDER_GD)
          if ( (err = gdImageStringTTF (0, brect, 0, FIXCARG fn, fontsize,0.0,0,0,"Hello") ) == 0 )
#else
          if ( (err = gdImageStringFT (0, brect, 0, FIXCARG fn, fontsize,0.0,0,0,"Hello") ) == 0 )
#endif
          {
            do_truetype = 1;
            fontfile = fn;
            if (do_debug)
              Tcl_AppendResult(interp, "{DEBUG: found as ", fontfile, "}\n", 0);
          }
          else
          {
            if (do_debug)
              Tcl_AppendResult(interp, "{DEBUG: not found\n", "Truetype error: ", err, "}", 0);
          }
#endif
        }
      }
      else if ( strcmp(argv[0], "-single") == 0 )
      {
        singleline = 1;
      }
      else if ( strcmp(argv[0], "-antialias") == 0 )
        doantialias = 1;
      else if ( strcmp(argv[0], "-noantialias") == 0 )
        doantialias = -1;
      else if ( strcmp(argv[0], "-render") == 0 )
        do_render = 1;
      else if ( strcmp(argv[0], "-norender") == 0 )
        do_render = 0;

      argc--;
      argv++;
    }

    if (string == 0 )
    {
      Tcl_SetResult(interp, usage_message, TCL_STATIC);
      return TCL_ERROR;
    }
    
    /* Set the format flags */

    /* Calculate the rectangle */

    /* Adjust the rectangle according to the anchor */

    /* Print the text */
    if (do_truetype)
    {
#if defined(ENABLE_TRUETYPE)
        
      wid = width;
      hgt = 20000;
      /* Replace all newlines with blanks if singleline is requested */
      if ( singleline )
      {
        char *cp;
        char *stripouts = "\r\n\f\t";
        char *cur;

        for ( cur = stripouts; *cur != '\0'; cur++ )
        {
          cp = strchr( (char *)string, *cur );
          while (cp)
          {
            *cp = ' ';
            cp = strchr(cp+1, *cur);
          }
        }
               
        wid = 32767;
      }

      /* The case of newlines is taken care of in the gdImageStringTTF (or gdImageStringTTFEx) call */
      if (do_render == 0 )
        im = 0;

      gdTextBlockTTF( interp, im, fontfile, fontsize, x, y, doantialias*color, do_debug,
                      FIXCARG string, anchorstyle, justifystyle,
                      DEGTORAD(angle), &wid, &hgt);

#endif
    }
    else
    {
      wid = width;
      hgt = 20000;

      gdTextBlockFixed(interp, im, font, x, y, color, do_debug,
                       string, anchorstyle, justifystyle, DEGTORAD(angle), &wid, &hgt);
        
    }

      
    {
      char msgstr[128];
      sprintf(msgstr,
              "{ height %d } { width %d }", hgt, wid);
      Tcl_AppendResult(interp, msgstr, 0);
    }

    if (do_debug)
    {
      char message[4095];
      sprintf(message, "Text: Used font %s at (%d,%d) with size %f angle %f orientation %d\n",
             fontname?fontname:"No font", x, y, fontsize, angle, orient);
      Tcl_AppendResult(interp, "\n{DEBUG: ", message, " }\n", 0);
    }
    return TCL_OK;
  }
  
  Tcl_SetResult(interp, usage_message, TCL_STATIC);
  return TCL_ERROR;
}


static int gdTextLineFixed(Tcl_Interp *interp,
                           gdImagePtr im, gdFontPtr font, 
                           int x0, int y0, 
                           int color, int debug,
                           const char *string, int len, double angle,
                           int *wid, int *hgt)
{
    
  string = gdSkipLeadingWhite(string);
    
  if (im != 0) 
  {
    double q = angle;
    while (q > DEGTORAD(360.0) )
      q -= DEGTORAD(360.0);
    while (q < 0.0)
      q += DEGTORAD(260.0);

    if ( (q > DEGTORAD(45.0) && q < DEGTORAD(135.0) ) || 
         (q > DEGTORAD(225.0) && q < DEGTORAD(315.0)) )
    {
      /* up */
      gdImageStringUpEx(im, font, x0, y0, FIXCARG string, color, 1);
    } else {
      /* normal */
      gdImageStringEx(im, font, x0, y0, FIXCARG string, color, 1);
    }
  }

  *wid = font->w * len;
  *hgt = font->h;

  return 1;
}
                           
static int gdTextLineTTF(Tcl_Interp *interp,
                           gdImagePtr im, const char *fontfile, double fontsize, 
                           int x0, int y0, 
                           int color, int debug,
                           const char *string, int len, double angle,
                           int *wid, int *hgt)
{
#if defined(ENABLE_TRUETYPE)
  char *err;
  int brect[8];
  int skip = 0;
  
  brect[0] = brect[1] = brect[2] = brect[3] = brect[4] = brect[5] = brect[6] = brect[7] = 0;

  /* Strip off any leading white space */
  {  
    const char *cp;
    cp = gdSkipLeadingWhite(string);
    skip = cp - string;  
  }

  #if defined(OLDER_GD)
    if ( (err = gdImageStringTTF (im, brect, color, 
                                  FIXCARG fontfile, fontsize, angle, x0, y0, FIXCARG string + skip) )
      Tcl_AppendResult(interp, "{ERROR: ", err, "}\n", 0);
  #else
    if ( (err = gdImageStringFT (im, brect, color, 
                                  FIXCARG fontfile, fontsize, angle, x0, y0, FIXCARG string + skip) ) )
      Tcl_AppendResult(interp, "{ERROR: ", err, "}\n", 0);
  #endif
  /* The following calculation may be version dependent. It assumes the following order
  ** in brect, and x in the even slots, y in the odd slots
  ** UL3 UR2
  ** LL0 LR1
  */
  if ( debug > 0 )
  {
    char ic[13];
    int k;
    Tcl_AppendResult(interp, "{DEBUG: For string |", string + skip, "| brect = ", 0);
    for (k=0;k<8;k++)
    {
      sprintf(ic, "%d", brect[k]);
      Tcl_AppendResult(interp, ic, ", ", 0);
    }
  }
  
  *wid = brect[2] - brect[0];
  *hgt = brect[1] - brect[7];

    
  if ( debug > 0 )
  { 
    char ic[13];
    sprintf(ic, "%d", *hgt);
    Tcl_AppendResult(interp, "height=", ic, " ", 0);
    Tcl_AppendResult(interp, "}\n", 0);
  } 
#endif
  return 1;
}
                           
static int gdTextBlockFixed(Tcl_Interp *interp,
                           gdImagePtr im, gdFontPtr font, 
                           int x0, int y0, 
                           int color, int debug,
                           const char *string, int anchorflags, int justifyflags, double angle,
                           int *wid, int *hgt)
{
  /* For now, block all text left. Later, use the flags to anchor, justify as needed */
  /* If the height is <=0 or width <= 0, assume it's a single line */
  char *copy;
  int place = 0;
  int len;
  int slen;
  int iwid = *wid;
  int ihgt = *hgt;
  char *estring = (char *)string;
  int x = x0;
  int y = y0;
  int interchar = 0; /* This is non-zero for full line justification: not implemented */
    
  /* We must make a copy after converting from UTF */
#if TCL_MAJOR_VERSION > 8 || (TCL_MAJOR_VERSION == 8 && TCL_MINOR_VERSION >=1)
  Tcl_DString tds;

  Tcl_DStringInit(&tds);
  Tcl_UtfToExternalDString(NULL, string, -1, &tds);
  estring = Tcl_DStringValue(&tds);
#endif

  slen = strlen(estring);
  
  *hgt = 0;
  *wid = 0;
  if ( (copy = Tcl_Alloc(slen+1)) == 0 )
  {
    return 0;
  }
    
  if ( im != 0 ) /* to prevent recursion */
  {
    int ww=iwid, hh=ihgt;
    char buffer[256];
    gdTextBlockFixed(interp, 0, font, x0, y0, 
                     color, debug, string, anchorflags, justifyflags, 
                     angle, &ww, &hh);
    /* At this point, enough is known to adjust the height, as well as enough
     * to do horizontal justification. */
    /* Adjust the y */
    if (debug > 0)
    {      
      sprintf(buffer, "width=%d, height=%d", ww, hh);
      Tcl_AppendResult(interp, "{DEBUG: Anchor calculation: ", buffer, "}\n", 0);
    }
    if ( anchorflags == TK_ANCHOR_W || anchorflags == TK_ANCHOR_CENTER || anchorflags == TK_ANCHOR_E )
      y -= (hh+1) / 2;
    else if ( anchorflags == TK_ANCHOR_SW || anchorflags == TK_ANCHOR_S || anchorflags == TK_ANCHOR_SE )
      y -= hh;

    if ( anchorflags == TK_ANCHOR_CENTER || anchorflags == TK_ANCHOR_N || anchorflags == TK_ANCHOR_S )
      x -= (ww+1) / 2;
    else if (anchorflags == TK_ANCHOR_NE || anchorflags == TK_ANCHOR_E || anchorflags == TK_ANCHOR_SE )
      x -= ww;
      
    /* At this point, x and y represent the upper left corner of the text block */
    iwid = ww;  
  }
    
  {
    int jwid;
    int jhgt;
    int count=0;
      
    while (place < slen && count < 100)
    {
      int w = iwid;
      int myx = x;
        
      count++;
      jwid = iwid;
      jhgt = ihgt;
      len = gdFitWidthFixed(estring+place, w, font);

      if (len <= 0)
        len = slen - place;
      strncpy(copy, estring+place, len);
      copy[len] = '\0';
        
      /* For a fixed-width font, the actual width is a multiple of the length */
      w = strtrim(copy) * font->w;

      /* Fix x by considering the effects of anchor placement */

      if ( justifyflags == TK_JUSTIFY_CENTER )
          myx += (iwid - w) / 2;
      else if (justifyflags == TK_JUSTIFY_RIGHT)
          myx += iwid - w ;
        
      gdTextLineFixed(interp, im, font, myx, y, color, debug, copy, len, angle, &jwid, &jhgt);

      if ( *wid < w )
        *wid = w;
      *hgt += jhgt;
      y += jhgt;
      place += len;
    }
  }
  Tcl_Free(copy);
#if TCL_MAJOR_VERSION > 8 || ( TCL_MAJOR_VERSION == 8 && TCL_MINOR_VERSION >= 1)
  Tcl_DStringFree(&tds);
#endif
  return 1;
}

/****************************************************************
 * The algorithm for Text Block is as follows:
 * 1) Figure out the dimensions of the text block
 * 2) Adjust the horizontal center appropriately
 * 3) Go through the lines one at a time
 *    Adjust each line's justification
 * The hard part is that the dimensions of the text block must
 * be figured out before the rest of the algorithm makes sense.
 * This sounds like a problem in recursion.
 * Perhaps the best way to work it is to build up a pile of 
 * structures, one representing each output line.
 ****************************************************************/
struct LineRec
{
    int width;
    char *string;
    struct LineRec *next;
};

static struct LineRec *newLineRec(void)
{
    struct LineRec *retval;
    retval = (struct LineRec *)Tcl_Alloc(sizeof (struct LineRec));
    return retval;
}

static struct LineRec *createLineRec(const char *str, int w)
{
    struct LineRec *retval = newLineRec();
    if (retval)
    {
        retval->width = w;
        retval->next = 0;
        retval->string = Tcl_Alloc(strlen(str) + 1);
        if ( retval->string )
            strcpy(retval->string, str);
    }
    return retval;
}

static void freeLineRec (struct LineRec *lr)
{
    if (lr && lr->string)
        Tcl_Free(lr->string);
    if (lr)
        Tcl_Free((char *)lr);
}

static void freeLineRecs (struct LineRec *lr )
{
    struct LineRec *n;
    while (lr)
    {
        n = lr->next;
        freeLineRec(lr);
        lr = n;
    }
}

static int gdRenderLineRecs(Tcl_Interp *interp,
                            gdImagePtr im, const char *fontfile, double fontsize,
                            int x0, int y0,
                            int color, int debug,
                            struct LineRec *head,
                            int justifyflags,
                            double angle,
                            int width,
                            double lineSpacing)
{
    struct LineRec *current;
    int numlines = 0;
    
    for (current = head; current ; current = current->next )
    {
        int myx = x0;
        int myy = y0 + (int)(numlines * lineSpacing);
        int w, h;
        
        if ( justifyflags == TK_JUSTIFY_CENTER )
            myx += ( width - current->width) / 2;
        else if ( justifyflags == TK_JUSTIFY_RIGHT)
            myx += ( width - current->width );
        
        gdTextLineTTF(interp, im, fontfile, fontsize, myx, myy, color, debug, 
                      current->string, strlen(current->string), angle, &w, &h);
        
        numlines++;
    }
    
    return TCL_OK;
}
                            
static int gdTextBlockTTF(Tcl_Interp *interp,
                           gdImagePtr im, const char *fontfile, double fontsize, 
                           int x0, int y0, 
                           int color, int debug,
                           const char *string, 
                           int anchorflags, int justifyflags,
                           double angle,
                           int *wid, int *hgt)
{
  /* For now, block all text left. Later, use the flags to anchor, justify as needed */
  /* If the height is <=0 or width <= 0, assume it's a single line */
  char *copy;
  int place = 0;
  int len;
  int slen = strlen(string);
  int iwid = *wid;
  int ihgt = *hgt;

  int x = x0;
  int y = y0;
  int interchar = 0; /* This is non-zero for full line justification purposes: not implemented */ 
  double interLine = fontsize;  
    
  struct LineRec *head = 0;
  int maxwidth = 0;
  int maxheight = 0;
    
  /* Since truetype anchors are in the lower left, and the UL is needed, 
   * subtract off a fontsize
   */
  y = (int)(y+fontsize);
    
  if ( (copy = Tcl_Alloc(slen+1)) == 0 )
  {
    *wid = 0;
    *hgt = 0;
    return 0;
  }
  *hgt = 0;
  *wid = 0;
    
  if ( debug > 0 )
    Tcl_AppendResult(interp, "{DEBUG: Entering BlockTTF}\n", 0);
    
  /* I know this is idiotic, but I need the height of the font, not the control box */
  {
    int w=0;
    int h=0;

    gdTextLineTTF(interp, 0, fontfile, fontsize, x0, y, color, debug, "Tyekl;", 2, 0.0, &w, &h);
    interLine = h ;
  }
    
  /* Let's start breaking the string apart into LineRecs */
  {
    struct LineRec *cur = 0;
    
    while ( place < slen )
    {
      int w = iwid;
      struct LineRec *n;

      len = gdFitWidthTTF(interp, string+place, &w, fontfile, fontsize );
      if ( len <= 0 )
        len = slen - place;
      strncpy (copy, string+place, len);
      copy[len] = '\0';
      strtrim(copy);
            
      if ( (n = createLineRec(copy, w)) != 0 )
      {
        if ( head == 0 )
          head = n;
        else
          cur->next = n;
        cur = n;
      }
      place += len;
    }
        
  }
    
  /* Let's get the width right */
  {
      struct LineRec *cur;
      int linecount = 0;
      for ( cur = head; cur; cur = cur->next)
      {
        if ( cur->width > maxwidth )
            maxwidth = cur->width;
        linecount++;
        if ( debug > 0 )
        {
          char buf[128];
          sprintf(buf, "Line %d, width %d", linecount+1, cur->width);
          Tcl_AppendResult(interp, "{DEBUG: Line Records: ", buf, " text |", cur->string, "|}\n", 0);
        }
      }
      maxheight = (int)(linecount*interLine);

  }
    
  /* Now, let's get the anchoring right */
  if ( anchorflags == TK_ANCHOR_W || anchorflags == TK_ANCHOR_CENTER || anchorflags == TK_ANCHOR_E )
    y -= (maxheight+1) / 2;
  else if ( anchorflags == TK_ANCHOR_SW || anchorflags == TK_ANCHOR_S || anchorflags == TK_ANCHOR_SE )
    y -= maxheight;

  if ( anchorflags == TK_ANCHOR_CENTER || anchorflags == TK_ANCHOR_N || anchorflags == TK_ANCHOR_S )
    x -= (maxwidth+1) / 2 ;
  else if (anchorflags == TK_ANCHOR_NE || anchorflags == TK_ANCHOR_E || anchorflags == TK_ANCHOR_SE )
    x -= maxwidth;

  *wid = maxwidth;
  *hgt = maxheight;

  /* Now, let's get the LineRecs rendered */
  gdRenderLineRecs(interp,im,fontfile,fontsize,
                   x, y,                        /* These must be adjusted for anchoring */
                   color, debug, head, justifyflags, angle,
                   maxwidth,                      /* This must be adjusted for the "real" width */
                   interLine);
    
  /* Now, we must free up the LineRecs */
  freeLineRecs(head);  
  head = 0;

  return TCL_OK;
}

/* This function is not Unicode aware at this time.
** What is the equivalent in unicode?
*/
static int gdSingleLine(const char *init, char *line, int len)
{
  const char *j = init;
  char *k = line;
  int l = 0;
  int extra = 0;
  
  /* Skip initial white space */
  j = gdSkipLeadingWhite(init);
  extra = j - init;  
  
  /* Copy characters until len or another break character is found */
  while (l < len && *j && (*j != '\r' && *j != '\n' && *j != '\f') )
    k[l++] = *j++;
  k[l++] = '\0';
    
  /* OK -- Now "trim" at the right */
  while (l > 0 && ( k[l-1] == ' ' || k[l-1] == '\t'))
    k[--l] = '\0';

  len = l;
  return j - init;
}

/* Returns the index of the character to break at */
/* May want to strip white space out of this arrangement */
static int gdFitWidthFixed(const char *init, int width, gdFontPtr font)
{
  int lens;
  int lenw;
  int lenb;
  const char *cp;
  int min;
  int extra=0;
    
  /* Skip over initial white space */
  cp = gdSkipLeadingWhite(init);
  extra = cp - init;  
    
  /* Get string length remaining -- start by assuming everything goes */
  lens = strlen(cp);
  min = lens;
    
  /* If there is a break character, get even less */
  lenb = gdFirstBreak(cp);
  if (lenb != -1 && lenb < min)
    min = lenb;
    
  /* If there isn't enough room, get a linewidth's worth */
  lenw = width / font->w ;
  if ( width >= 0 && lenw < min)
  {
    min = lenw;
    /* If the line breaks a word, back up if possible */
    if ( cp[lenw] != ' ' && cp[lenw] != '\t' && 
         (lenw > 1 && cp[lenw-1] != ' ') )
    {
      int lenww;
      lenww = gdLastSpaceBefore(cp, lenw);
      if (lenww != -1)
        min = lenww;      
    }
  }
    
  return min+extra;
}

/* Returns the index of the character to break at
** Assumes font exists
*/
static int gdFitWidthTTF(Tcl_Interp *interp,
                         const char *init, int *width, 
                         const char *fontfile, double fontsize)
{
  char *string;
  int retval = 0;
  int rlen = 1;
  int i;
  int len;
  int ilen = strlen(init);
  int slen;
  int retwidth = 0;
  int wbreak = 0, wbreaklen=0;

  if ( (string = Tcl_Alloc(ilen+1)) == 0 )
    return 0;
    
  for (i=1; i <= ilen; i++)
  {
    int wid = *width;
    int hgt;
    /* Calling gdSingleLine results in a string which is stripped and whose
     * offsets do _not_ correspond to the original string */
    rlen = gdSingleLine(init, string, i);
    slen = strlen(string);      
      
    if ( wid <= 0 )
      wid = 32767;

    /* No debug -- but provide the interp anyway */
    len =  gdTextLineTTF(interp, 0, fontfile, fontsize,
                         0, 0, 0, 0, string, slen, 0.0, &wid, &hgt);

    if ( *width == 0 || wid <= *width )
    {
      retval = rlen;
      retwidth = wid;
    }
    else
    {
      char buff[200];
      sprintf(buff, "retval=%d; rlen=%d; wid=%d; width=%d", retval, rlen, wid, *width);
      Tcl_AppendResult(interp,"{DEBUG: Width exceeded for: |", string, "| at: ", buff, "}\n", 0);
      break;
    }  

    /* If an "absolute" break is reached, output the current line */
    if ( init[rlen] == '\r' || init[rlen] == '\n' || init[rlen] == '\f' || init[rlen] == '\0' )
    {
      char buff[200];
      sprintf(buff, "rlen=%d, char 0x%x", rlen, init[rlen]);
      Tcl_AppendResult(interp,"{DEBUG: Absolute break at: ", buff, "}\n", 0);
      break;
    }
    
    /* If the last character is a whitespace, remember where we are */
    if ( init[rlen] == ' ' || init[rlen] == '\t')
    {
      /*
      char buf[256];
      sprintf(buf, "%d (%d)", retval, retwidth);
      Tcl_AppendResult(interp, "{DEBUG: Marked break at position ", buf, " for string |", init, "|}\n", 0);
      */
      wbreak = retwidth;
      wbreaklen = retval;
    }
  }

  Tcl_Free(string);
  if ( wbreak != 0 && retval != wbreaklen && retval != rlen && 
       init[retval] != '\0' && init[retval] != ' '  && init[retval] != '\t' && 
       init[retval] != '\r' && init[retval] != '\n' && init[retval] != '\f' )
  {
    char buf[256];
    sprintf(buf, "retval=%d; retwidth=%d, wbreak=%d, wbreaklen=%d", retval, retwidth, wbreak, wbreaklen);
    Tcl_AppendResult(interp, "{DEBUG: Updating break: ", buf, " }\n", 0);
      
    retwidth = wbreak;
    retval = wbreaklen;
  }
    
  /*
   {
     char buf[256];
     sprintf(buf, "width=%d; length=%d", retwidth, retval);
     Tcl_AppendResult(interp, "{DEBUG: Returning ", buf, " }\n", 0);
   }
  */
  *width = retwidth;  
  return retval;
}

#if defined(ENABLE_TRUETYPE)
#include <freetype.h>
static int GDFontAliases (ClientData unused, Tcl_Interp *interp, int argc, char **argv)
{
    static char usage[] = "gd fontalias globspec ...";
    /* Without changing gd-2.0.1, the "library" built there cannot be exposed.
     * Thus, I hope we can build our own "library" here and destroy it when done
     */
    int error;
    FT_Face face;
    FT_Library library;
    int i;
    
    Tcl_DString results;
    Tcl_DString globresult;
    Tcl_DString script;
    
    Tcl_DStringInit(&results);
    
    if ( (error=FT_Init_FreeType(&library)) != 0 )
    {
        Tcl_AppendResult(interp, "gd fontalias: Cannot initialize FreeType library", 0);
        return TCL_ERROR;
    }
    
    if ( argc < 1 )
    {
        Tcl_AppendResult(interp, "gd fontalias: You must specify at least one directory specification\n",
                         usage, 0);
        return TCL_ERROR;
    }
    
    for ( i = 0 ; i < argc; i++ )
    {
        /* Glob the filenames */
        Tcl_DStringInit(&script);
        Tcl_DStringAppend(&script, "glob ", -1);
        Tcl_DStringAppend(&script, argv[i], -1);

        if ( Tcl_GlobalEval(interp, Tcl_DStringValue(&script)) != TCL_ERROR )
        {
            int elc;
            char **elv;
            
            Tcl_DStringInit(&globresult);
            Tcl_DStringGetResult(interp, &globresult);
            if ( Tcl_SplitList(interp, Tcl_DStringValue(&globresult), &elc, &elv) != TCL_ERROR )
            {
                int j;
                for (j = 0; j < elc; j++ )
                {
                    /* for each filename, get the faces */
                    if ( FT_New_Face(library, elv[j], 0, &face) == 0 )
                    {
                        char facename[2048];
                        strcpy(facename, face->family_name);
                        if ( face->style_name != NULL )
                        {
                            strcat(facename, " ");
                            strcat(facename, face->style_name);
                        }
                        /* For each face, build the array spec */
                        Tcl_DStringAppendElement(&results, facename);
                        Tcl_DStringAppendElement(&results, elv[j]);
                        Tcl_DStringAppend(&results, "\n", -1); /* Just for readability */
                    }
                    FT_Done_Face(face);
                }
                Tcl_Free((char *)elv);
            }
            Tcl_DStringFree(&globresult);
        }
        else
        {
            /* The glob failed */
        }
        Tcl_DStringFree(&script);
    }
    
    Tcl_DStringResult(interp, &results);
    Tcl_DStringFree(&results); /* Now a different, reinited string */
    FT_Done_FreeType(library);
    return TCL_OK;
}
#endif

/****************************************************************
** This is the version information/command
** The static data should also be used by pkg_provide, etc.
****************************************************************/
/* Version information */
static char version_string[] = "0.4.0.3";

/* Version command */
static int Version(ClientData unused, Tcl_Interp *interp, int argc, char **argv)
{
  Tcl_SetResult(interp, version_string, TCL_STATIC);
  return TCL_OK;
}

/****************************************************************
** Debug procedure for gd library, so all errors are written to
** gd_tcl.err
****************************************************************/
#include <stdarg.h>
static void gd_tcl_extension_output (void *context, const char *fmt, va_list va)
{
  FILE *output = (FILE *)context;
  vfprintf(output, fmt, va);
  fflush(output);
}

/****************************************************************
** Initialization procedures
** These are the only public procedures in the file.
** These are OS independent
****************************************************************/
/* Initialization Procedures */
EXPORT(int,Gd_Init) (Tcl_Interp *interp)
{
  static int initialized = 0;
#if TCL_MAJOR_VERSION > 8 || (TCL_MAJOR_VERSION == 8 && TCL_MINOR_VERSION >= 1)
  #if defined(USE_TCL_STUBS)
    Tcl_InitStubs(interp, TCL_VERSION, 0 );
  #endif
#endif
    
#if TCL_MAJOR_VERSION > 8 || (TCL_MAJOR_VERSION == 8 && TCL_MINOR_VERSION >= 1)
  #if defined(USE_TK_STUBS)
    Tk_InitStubs(interp, TCL_VERSION, 0 );
  #endif
#endif

  if ( initialized == 0 )
  {
    init_image_names();
    init_font_aliases();
    initialized = 1;
  }
#if TCL_MAJOR_VERSION <= 7
  Tcl_CreateCommand(interp, "gd", gd, 
                    0, 0);
#else
  /* Wanted to use namespaces, but "unknown" isn't smart enough yet */
  /* Since this package is so full of numbers, this would be a great place
  ** to introduce a TclCmdObj
  */
  Tcl_CreateCommand(interp, "gd", gd, 
                    0, 0);
#endif

  Tcl_PkgProvide (interp, "gd", version_string);


#ifdef HAVE_STDARG_H
  /* Change the way errors are reported from GD */
  {
    char filename[4096+1];
    char *cp;
  
    filename[0] = '\0';
    if ( (cp = getenv("TEMP")) != 0 )
      strcpy (filename, cp);
    strcat(filename, "/gd_tcl.err");

#if 0
    if ( (fp = fopen(filename, "w")) != NULL )
      set_GD_debug_fn(gd_tcl_extension_output, (void *)fp);
#endif
  }
#endif

  return TCL_OK;
}

/* The gd function is considered safe. */
EXPORT (int,Gd_SafeInit) (Tcl_Interp *interp)
{
  return Gd_Init(interp);
}

/* Exported symbols */
BOOL APIENTRY DllEntryPoint (HINSTANCE hInstance, DWORD reason, LPVOID lpCmdLine)
{
  switch (reason)
  {
    case DLL_PROCESS_ATTACH:
      /* Moved initialization from here to avoid stubs problems */
      break;
    case DLL_THREAD_ATTACH:
      break;
    case DLL_PROCESS_DETACH:
      /* Release all the memory allocated by the images */
      delete_image_names();
      delete_font_aliases();
      break;
    case DLL_THREAD_DETACH:
      break;
  }
  /* Don't do anything, so just return true */
  return TRUE;
}

static Tcl_HashTable images;
static unsigned long image_count=0L;
static char image_name[7 + 11 + 1];

static void init_image_names(void)
{
  Tcl_InitHashTable(&images, TCL_STRING_KEYS);
}

static void delete_image_names(void)
{
  /* Only to be called when the DLL is detaching */
  Tcl_HashEntry *ent;
  Tcl_HashSearch srch;
  gdImagePtr im;

  for (ent = Tcl_FirstHashEntry(&images, &srch); ent != 0; 
       ent = Tcl_NextHashEntry(&srch) )
  {
    if  ( (im = (gdImagePtr)Tcl_GetHashValue(ent)) != 0)
      gdImageDestroy(im);
  }

  Tcl_DeleteHashTable(&images);
}

static const char *make_image_name(void)
{
  int status;
  sprintf(image_name, "gdImage%lu", image_count++);
  Tcl_CreateHashEntry(&images, image_name, &status);
  return image_name;
}

static int  set_image(const char *name, gdImagePtr im)
{
  Tcl_HashEntry *data;
  if ( (data = Tcl_FindHashEntry(&images, name)) != 0 )
  {
    Tcl_SetHashValue(data, im);
    return 1;
  }
  return 0;
}

static int  image_name_valid(const char *name)
{
  void *data;
  data = Tcl_FindHashEntry(&images, name);
  return (data != 0);
}

static int  image_valid(const char *name)
{
  Tcl_HashEntry *entr;
  ClientData data = 0;
  entr = Tcl_FindHashEntry(&images, name);
  if (entr != 0)
    data = Tcl_GetHashValue(entr);
  return data != 0;
}

static int  delete_image_name(const char *name)
{
  Tcl_HashEntry *data;
  data = Tcl_FindHashEntry(&images, name);
  if (data != 0)
    Tcl_DeleteHashEntry(data);
  return data != 0;  
}

static gdImagePtr get_image(const char *name)
{
  Tcl_HashEntry *data;
  if ( (data = Tcl_FindHashEntry(&images, name)) != 0 )
    return (gdImagePtr)Tcl_GetHashValue(data);
  return 0;
}

static const char *get_image_name (gdImagePtr im)
{
  Tcl_HashEntry *ent;
  Tcl_HashSearch srch;
  for (ent = Tcl_FirstHashEntry(&images, &srch); ent != NULL; 
       ent = Tcl_NextHashEntry(&srch) )
  {
    if ( ent->clientData == (ClientData)im )
      return ent->key.oneWordValue;
  }
  return 0;  
}

static int lArc(const char *name, int xul, int yul, int xlr, int ylr,
                                 int sangle,  int extent, int color, int thick)
{
  /* The plain old gdImageArc function */
  gdImagePtr im;
  int cx, cy, w, h;
  
  if ( (im = get_image(name) ) == 0 )
    return 0;

  cx = (xul + xlr) / 2;
  cy = (yul + ylr) / 2;
  w  = (xlr - xul);
  h  = (ylr - yul);
  
  gdImageArcEx(im, cx, cy, w, h, sangle, sangle + extent, color, thick);
  return 1;
}

static int lPie(const char *name, int xul, int yul, int xlr, int ylr,
                                 int sangle,  int extent, int color, int thick)
{
  /* gdImageArc followed by 2 lines to the center */
  gdImagePtr im;
  int cx, cy, w, h;
  int xr0, yr0, xr1, yr1;
  
  if ( (im = get_image(name) ) == 0 )
    return 0;

  cx = (xul + xlr) / 2;
  cy = (yul + ylr) / 2;
  w  = (xlr - xul);
  h  = (ylr - yul);
  xr0 = xr1 = cx;
  yr0 = yr1 = cy;
  xr0 += (int)(w * cos( DEGTORAD(sangle) ) / 2.0 );
  yr0 += (int)(h * sin( DEGTORAD(sangle) ) / 2.0 );
  xr1 += (int)(w * cos( DEGTORAD(sangle+extent) ) / 2.0 );
  yr1 += (int)(h * sin( DEGTORAD(sangle+extent) ) / 2.0 );
  
  
  gdImageArcEx(im, cx, cy, w, h, sangle, sangle + extent, color, thick);
  gdImageLineEx(im, cx, cy, xr0, yr0, color, thick);
  gdImageLineEx(im, cx, cy, xr1, yr1, color, thick);

  return 1;
}

static int lChord(const char *name, int xul, int yul, int xlr, int ylr,
                                 int sangle,  int extent, int color, int thick)
{
  /* gdImageArc followed by 1 line between the endpoints */
  gdImagePtr im;
  int cx, cy, w, h;
  int xr0, xr1, yr0, yr1;
  
  if ( (im = get_image(name) ) == 0 )
    return 0;

  cx = (xul + xlr) / 2;
  cy = (yul + ylr) / 2;
  w  = (xlr - xul);
  h  = (ylr - yul);
  xr0 = xr1 = cx;
  yr0 = yr1 = cy;
  xr0 += (int)(w * cos( DEGTORAD(sangle) ) / 2.0 );
  yr0 += (int)(h * sin( DEGTORAD(sangle) ) / 2.0 );
  xr1 += (int)(w * cos( DEGTORAD(sangle+extent) ) / 2.0 );
  yr1 += (int)(h * sin( DEGTORAD(sangle+extent) ) / 2.0 );
  
  
  gdImageArcEx(im, cx, cy, w, h, sangle, sangle + extent, color, thick);
  gdImageLineEx(im, xr0, yr0, xr1, yr1, color, thick);

  return 1;
}

/******************************************************************************
** GDOpen - Create a new image either from scratch or an image file.
******************************************************************************/
static int GDOpen      (ClientData unused, Tcl_Interp *interp, int argc, char **argv)
{
  static char usage[] = "gd open -width w -height h -file f -filetype ["
#ifdef ENABLE_GIF
                        "gif|"
#endif
#ifdef ENABLE_JPEG
                        "jpeg|"
#endif
                        "png|xbm] "
                        "-background color\n"
                        "\treturns token for image";
  gdImagePtr im, tmpim;
  int width = 0, height = 0;
  char *file = 0, *filetype = 0;
  char *colorname;
  int do_color=0;
  int i;

  for (i=0; i<argc; i++)
  {
    if ( strcmp(argv[i], "-width") == 0 )
      width = atoi(argv[++i]);
    else if ( strcmp(argv[i], "-height") == 0 )
      height = atoi(argv[++i]);
    else if ( strcmp(argv[i], "-file") == 0 )
      file = argv[++i];
    else if ( strcmp(argv[i], "-filetype") == 0 )
      filetype = argv[++i];
    else if ( strcmp(argv[i], "-background") == 0)
    {
      do_color = 1;
      colorname=argv[++i];
    }
    else
    {
      Tcl_SetResult(interp, usage, TCL_STATIC);
      return TCL_ERROR;
    }
  }

  /* Now what have we got? */
  if ( file != 0 )
  {
    FILE *fin;
    if ( (fin = fopen(file, "rb")) == NULL)
    {
      Tcl_AppendResult(interp, "File Error: Cannot open ", file, " for read", NULL);
      return TCL_ERROR;
    }
    
    if ( filetype == 0 )
    {
      filetype = "png";
      if ( strcmp ("xbm", file + strlen(file) - 3) == 0 )
        filetype = "xbm";
#ifdef ENABLE_GIF
      else if ( strcmp ( "gif", file + strlen(file) - 3 ) == 0 )
        filetype = "gif";
#endif
#ifdef ENABLE_JPEG
      else if ( strcmp ( "jpg", file + strlen(file) - 3 ) == 0 )
        filetype = "jpeg";
#endif
    }

    if ( strcmp(filetype, "xbm") == 0 )
      tmpim = gdImageCreateFromXbm(fin);
#ifdef ENABLE_GIF
    else if ( strcmp(filetype, "gif") == 0 )
      tmpim = gdImageCreateFromGif(fin);
#endif
#ifdef ENABLE_JPEG
    else if ( strcmp(filetype, "jpeg") == 0 )
      tmpim = gdImageCreateFromJpeg(fin);
#endif
    else
      tmpim = gdImageCreateFromPng(fin);
    
    if ( width <= 0 || height <= 0 ) /* Create FROM the file */
    {
      im = tmpim;
    }
    else  /* Create the image and copy the file image into it after resizing */
    {
      im = gdImageCreate(width, height);
      gdImageCopyResized (im, tmpim, 0, 0, 0, 0, width, height, 
                          gdImageSX(tmpim), gdImageSY(tmpim));
      gdImageDestroy(tmpim);
    }
    
    fclose (fin);
  }
  else if ( width <= 0 || height <= 0 )
  {
    Tcl_SetResult(interp, usage, TCL_STATIC);
    return TCL_ERROR;
  }
  else
  {
    im = gdImageCreate(width, height);
    if (im && do_color)
    {
      /* Must be a solid color! */
      if ( colorname && isdigit(colorname[0]) )
        color_select(im, colorname);
    }
  }

  if (im)
  {
    const char *imname;
    imname = make_image_name();
    if ( set_image(imname, im) == 1 )
      Tcl_SetResult(interp, (char *)imname, TCL_VOLATILE);
    else
    {
      Tcl_SetResult(interp, "Unable to add image to internal table", TCL_VOLATILE);
      return TCL_ERROR;
    }
  }
  else
  {
    Tcl_SetResult(interp, "Can't create image in gdImageCreate", TCL_STATIC);
    return TCL_ERROR;
  }
  
  return TCL_OK;
}

/******************************************************************************
** GDClose - Discard an image, allowing save if desired
**           Duplicates argument parsing in GDSave, but otherwise uses it to
**           do any necessary saves.
** Note: Will not discard image if save was requested and failed.
******************************************************************************/
static int GDClose     (ClientData unused, Tcl_Interp *interp, int argc, char **argv)
{
  static char usage[] = "gd close image "
                        "[-save [-file filename [-filetype "
#ifdef ENABLE_GIF
                        "gif|"
#endif
#ifdef ENABLE_JPEG
                        "jpeg|"
#endif
                        "png] ] ]";
  const char *name;
  char *filename = 0;
  char *filetype = 0;
  int  dosave    = 0;
  gdImagePtr im;
  int i;
  int status = TCL_OK;

  /* For the call to GDSave */
  int sargc = argc;
  char **sargv = argv;

  if (argc < 1)
  {
    Tcl_SetResult(interp, usage, TCL_STATIC);
    return TCL_ERROR;
  }

  name = argv[0];
  argc--;
  argv++;
  if ( (im = get_image(name)) == 0 )
  {
    Tcl_AppendResult(interp, "Cannot close: ", name, " is not a valid image name", NULL);
    return TCL_ERROR;
  }

  /* 
  ** Now look for other arguments.
  ** Do any argument checking here so any error messages come from this module.
  */
  for (i=0; i<argc; i++)
  {
    if ( strcmp(argv[i], "-save" ) == 0 )
    {
      dosave = 1;
    }
    else if ( strcmp(argv[i], "-file") == 0 )
    {
      dosave = 1;             /* If a file is specified, save is requested */
      filename = argv[++i];
    }
    else if ( strcmp(argv[i], "-filetype") == 0 )
    {
      filetype = argv[++i];
    }
  }

  /* Design question: If GDSave() fails, should we avoid destroying the image?
  ** For now assume yes, so a catch could save the image in a different location... 
  */
  if (dosave) 
    status = GDSave (unused, interp, sargc, sargv);
  
  /* Can't use this image again */
  if ( status == TCL_OK)
  {
    gdImageDestroy(im);
    delete_image_name(name);
  }
  
  return status;
}

/******************************************************************************
** GDSave - Save an image file.
**          No backups, checkpoints, or other fancy stuff at this time
******************************************************************************/
static int GDSave    (ClientData unused, Tcl_Interp *interp, int argc, char **argv)
{
  static char usage[] = "gd save image "
                        "[-file filename [-filetype "
#ifdef ENABLE_GIF
                        "gif|"
#endif
#ifdef ENABLE_JPEG
                        "jpeg|"
#endif
                        "png ] ]";
  const char *name;
  char *filename = 0;
  char *filetype = 0;
  gdImagePtr im;
  int i;
  FILE *fout;

  if (argc < 1)
  {
    Tcl_SetResult(interp, usage, TCL_STATIC);
    return TCL_ERROR;
  }

  name = argv[0];
  argc--;
  argv++;
  if ( (im = get_image(name)) == 0 )
  {
    Tcl_AppendResult(interp, "Cannot save: ", name, " is not a valid image name", NULL);
    return TCL_ERROR;
  }

  /* Now look for other arguments */
  for (i=0; i<argc; i++)
  {
    if ( strcmp(argv[i], "-file") == 0 )
      filename = argv[++i];
    else if ( strcmp(argv[i], "-filetype") == 0 )
      filetype = argv[++i];
    else if ( strcmp(argv[i], "-save") == 0 )
      ; /* Ignore this argument so that we can call this from GDClose */
  }
    
  if ( filename == 0 ) /* No filename specified...use image */
    filename = "image.png";

  if ( (fout = fopen(filename, "wb")) == 0 )
  {
    Tcl_AppendResult(interp, "File error: Can't open ", filename, " for writing", NULL);
    return TCL_ERROR;
  }

  /* Check the colormap. If nothing allocated, at least do a half-hearted job */
  if ( gdImageColorsTotal(im) == 0 )
  {
    /* Since we have no colors allocated, let's allocate 2 for starters */
    gdImageColorAllocate(im, 255, 255, 255);
    gdImageColorAllocate(im, 0, 0, 0);
  }
    
  if ( filetype == 0 ) /* No filetype specified--guess by the last 3 characters */
  {
    filetype = "png";
#ifdef ENABLE_GIF
    if ( strcmp ("gif", filename + strlen(filename) - 3) == 0 )
      filetype = "gif";
#endif
#ifdef ENABLE_JPEG
    if ( strcmp ("jpg", filename + strlen(filename) - 3) == 0 )
      filetype = "jpeg";
#endif
  }

  switch ( filetype[0] )
  {
#ifdef ENABLE_GIF
    case 'g': case 'G':
      gdImageGif(im, fout);
      break;
#endif
#ifdef ENABLE_JPEG
    case 'j': case 'J':
      gdImageJpeg(im, fout, -1); /* One day the quality parameter should be provided */
      break;
#endif
    case 'p': case 'P':
    default:
      gdImagePng(im, fout);
      break;
  }
  fclose(fout);
    
  return TCL_OK;
}

/******************************************************************************
** GDConfig - Report everything interesting about an image
**            Allow setting brush, tile, style, interlace, and transparent color index.
******************************************************************************/
static int GDConfig    (ClientData unused, Tcl_Interp *interp, int argc, char **argv)
{
  static char usage[] = "gd config image "
                        "[-brush image] "
                        "[-tile  image] "
                        "[-style int_list] "
                        "[-interlace 0|1] "
                        "[-transparent color] "
                        "[-colormap] "
                        "[-size] "
                        "[-fonttable arrayvar] "
                        ;

  const char *name;
  gdImagePtr tmpptr;
  gdImagePtr im;
  int i;
  const char *arrayname;  
  
  struct {
    unsigned style: 1;
    unsigned transparent: 1;
    unsigned brush: 1;
    unsigned tile : 1;
    unsigned interlace: 1;
    unsigned colormap:  1;
    unsigned size    :  1;
    unsigned fonttable: 1;  
  } report;
  
  if (argc < 1)
  {
    Tcl_SetResult(interp, usage, TCL_STATIC);
    return TCL_ERROR;
  }

  name = argv[0];
  argc--;
  argv++;
  if ( (im = get_image(name)) == 0 )
  {
    Tcl_AppendResult(interp, "Cannot config: ", name, " is not a valid image name", NULL);
    return TCL_ERROR;
  }  

  if ( argc == 0 )
  { 
    report.style = 1;
    report.transparent = 1;
    report.brush = 1;
    report.tile = 1;
    report.interlace = 1;
    report.colormap  = 1;
    report.size      = 1;
  }
  else
  {
    report.style = 0;
    report.transparent = 0;
    report.brush = 0;
    report.tile = 0;
    report.interlace = 0;
    report.colormap  = 0;
    report.size      = 0;
    
    for (i=0; i<argc; i++)
    {
      if ( strcmp(argv[i], "-brush") == 0 )
      {
        report.brush = 1;
        if ( argv[i+1] && argv[i+1][0] != '-' )
        {
          if ( (tmpptr  = get_image(argv[++i]) ) == 0 )
          {
            Tcl_AppendResult(interp, "Cannot locate image: ", argv[i], " name not valid", NULL);
            return TCL_ERROR;
          }
          else
          {
            gdImageSetBrush(im, tmpptr);
          }
        }
      }
      else if ( strcmp(argv[i], "-tile" ) == 0 )
      {
        report.tile = 1;
        if ( argv[i+1] && argv[i+1][0] != '-' )
        {
          if ( (tmpptr  = get_image(argv[++i]) ) == 0 )
          {
            Tcl_AppendResult(interp, "Cannot locate image: ", argv[i], " name not valid", NULL);
            return TCL_ERROR;
          }
          else
          {
            gdImageSetTile(im, tmpptr);
          }
        }
      }
      else if ( strcmp(argv[i], "-style" ) == 0 )
      {
        report.style = 1;
        if ( argv[i+1] && argv[i+1][0] != '-' )
        {
          /* Could allocate this, but for now let's just limit to 50 */
          int stylearr[50];
          int curcount = 0;
          char *ints = argv[++i];
          char *endstr = ints;

          for ( stylearr[0] = strtol(ints, &endstr, 0); 
                endstr != ints && curcount < 50 ; 
                ints = endstr, stylearr[curcount] = strtol(ints, &endstr, 0) )
            curcount++;
          if ( curcount > 0 )
            gdImageSetStyle(im, stylearr, curcount);
        }
      }
      else if ( strcmp(argv[i], "-interlace" ) == 0 )
      {
        report.interlace = 1;
        if ( argv[i+1] && argv[i+1][0] != '-' )
        {
          i++;
          if ( argv[i][0] == 'f' || atoi(argv[i]) == 0 )
            gdImageInterlace(im, 0);
          else
            gdImageInterlace(im, 1);
        }
      }
      else if ( strcmp(argv[i], "-transparent" ) == 0 )
      {
        int colind = 0;
        report.transparent = 1;
        if ( argv[i+1] && argv[i+1][0] != '-' )
        {
          int r, g, b;
          int status;

          r = 0;
          i++;
          if ( (status = sscanf(argv[i], "%d%d%d", &r, &g, &b)) == 3 )
            colind = color_index(im, r, g, b);
          else if ( (status = sscanf(argv[i], "#%d", &r) ) == 1 )
            colind = r;
          gdImageColorTransparent(im, colind);
        }
      }
      else if ( strcmp(argv[i], "-colormap" ) == 0 )
      {
        report.colormap = 1;
      }
      else if ( strcmp(argv[i], "-size" ) == 0 )
      {
        report.size = 1;
      }
      else if ( strcmp(argv[i], "-fonttable" ) == 0 )
      {
        if ( argv[i+1] && argv[i+1][0] != '-' )
        {
          arrayname = argv[++i];
          gdFontSetArray(interp, arrayname);

        }
        report.fonttable = 1;
      }
    }
  }

  if ( report.style )
  {
    int j;
    char num[10+1+1];
    Tcl_DString stylelist;
    Tcl_DStringInit(&stylelist);
    Tcl_DStringAppendElement(&stylelist, "-style");
    Tcl_DStringStartSublist(&stylelist);
    for (j=0; j< im->styleLength; j++)
    {
      sprintf(num, "%d", im->style[j]);
      Tcl_DStringAppendElement(&stylelist, num);
    }
    Tcl_DStringEndSublist(&stylelist);
    
    Tcl_AppendElement(interp, Tcl_DStringValue(&stylelist));
    Tcl_DStringFree(&stylelist);
  }
  if ( report.transparent )
  {
    sprintf(msgbuf, "-transparent %d", gdImageGetTransparent(im));
    Tcl_AppendElement(interp, msgbuf);
  }
  if ( report.brush )
  {
    const char *cp;
    if ( (cp = get_image_name(im->brush)) == 0 )
      cp = "Unknown";
    sprintf(msgbuf, "-brush %s", cp);
    Tcl_AppendElement(interp, msgbuf);
  }
  if ( report.tile )
  {
    const char *cp;
    if (im->tile == 0)
      cp = "{}";
    else if ( (cp = get_image_name(im->tile)) == 0 )
      cp = "Unknown";
    sprintf(msgbuf, "-tile %s", cp);
    Tcl_AppendElement(interp, msgbuf);
  }
  if ( report.interlace )
  {
    sprintf(msgbuf, "-interlace %s", gdImageGetInterlaced(im)?"true":"false");
    Tcl_AppendElement(interp, msgbuf);
  }
  if ( report.colormap )
  {
    int k;
    int kmax = gdImageColorsTotal(im);
    Tcl_DString colormap;
    Tcl_DStringInit(&colormap);
    Tcl_DStringAppendElement(&colormap, "-colormap");
    Tcl_DStringStartSublist(&colormap);           /* Sublist for the colormap */
    for (k = 0; k < kmax; k++ )
    {
      char num[10+1+1];

      if ( im->open[k] )                        /* No function to help with this bit... */
        continue;
        
      sprintf(num, "%d", k);
      Tcl_DStringStartSublist(&colormap);         /* Sublist for each entry */
      Tcl_DStringAppendElement(&colormap, num);
      Tcl_DStringStartSublist(&colormap);         /* Sublist for the color part */
      sprintf(num, "%d", gdImageRed(im, k));
      Tcl_DStringAppendElement(&colormap, num);
      sprintf(num, "%d", gdImageGreen(im, k));
      Tcl_DStringAppendElement(&colormap, num);
      sprintf(num, "%d", gdImageBlue(im, k));
      Tcl_DStringAppendElement(&colormap, num);
      Tcl_DStringEndSublist(&colormap);           /* Done with the color part */
      Tcl_DStringEndSublist(&colormap);           /* Done with the entry */
    }
    Tcl_DStringEndSublist(&colormap);             /* Done with the whole colormap */
    
    Tcl_AppendElement(interp, Tcl_DStringValue(&colormap));
    Tcl_DStringFree(&colormap);
  }
  if ( report.size )
  {
    sprintf(msgbuf, "-size { %d %d }", gdImageSX(im), gdImageSY(im));
    Tcl_AppendElement(interp, msgbuf);
  }
    
  if ( report.fonttable )
  {
    if (arrayname)
      gdFontGetArray(interp, arrayname);
  }
    
  return TCL_OK;
}

/*****************************************************************************
** color_index - Used by color_select to get RGB colors
** Strategy:
**  ALWAYS return a valid color
**  ALWAYS try to allocate a new color (for now) if there is no exact match,
**  unless there are no more.
**  THEN try to get a close match.
*****************************************************************************/
static int color_index(gdImagePtr im, int r, int g, int b)
{
  int i, imax;
  imax = gdImageColorsTotal(im);

  if ( (i = gdImageColorExact(im, r, g, b)) != -1 )
    return i;

  if ( imax < gdMaxColors )
    return gdImageColorAllocate(im, r, g, b);

  return gdImageColorClosest(im, r, g, b);
}

/******************************************************************************
** color_select is about the most complicated function. It works like this:
** i j k      - Color RGB. Reuse a color with these values if possible
**              If no colors can be allocated, return the closest allocated color
** #HHHHHH    - Hexidecimal specification of color (as on the canvas)
** &n         - Color index n directly
** =i1 i2 ... - Dashed line specification. return Styled. Values are color indices.
** @name      - Image name to use for brush. If name is blank, use 
**              brush already configured
** %name      - Image name to use for tile. If name is blank, use
**              tile already configured.
** If all else fails, return '0', the background color.
**
** But thus color_select will ALWAYS return a valid color.
******************************************************************************/
static int color_select (gdImagePtr im, const char *desc)
{
  int r, g, b;
  int color = 0;
  gdImagePtr tmpim;
  int style[50];  /* Artificial maximum */
  int istyle = 0;
  char *endstr;
  const char *ints;
  
  /*
  ** The color selection mechanism is so common, it seems best
  ** to set it up in one place.
  ** The algorithm is:
  **   If a series of 3 numbers is provided, they are rgb values.
  **   If one number is provided, it is a color index.
  **   If a name is provided that is a valid image name, it is a brush.
  */
  if ( sscanf(desc, "%d%d%d", &r, &g, &b) == 3 )
    color = color_index(im, r, g, b);
  else if ( sscanf(desc, "#%2.2x%2.2x%2.2x", &r, &b, &g) == 3 )
    color = color_index(im, r, g, b);
  else if ( sscanf(desc, "&%d", &r) == 1 )
  {
    int totcolors = gdImageColorsTotal(im);
    if ( totcolors > 0 )
      color = r % totcolors;
    else  /* NO colors at all! Allocate one and use it */
      color = gdImageColorAllocate(im, 255, 255, 255);
  }
  else if ( desc[0] == '=' )
  {
    ints = desc+1;
    for ( style[0] = strtol(ints, &endstr, 0); 
          endstr != ints && istyle < 50; 
          ints = endstr, style[istyle] = strtol(ints, &endstr, 0) )
            istyle++;
    if (istyle > 0 )
      gdImageSetStyle(im, style, istyle);
    color = gdStyled;    
  }
  else if ( desc[0] == '@' ) /* Brush */
  {
    if ( desc[1] != '\0' && (tmpim = get_image(desc+1)) != 0 )
    {
      gdImageSetBrush(im, tmpim);
      color = gdBrushed;
    }
  }
  else if ( desc[0] == '%' ) /* Tile */
  {
    if ( desc[1] != '\0' && (tmpim = get_image(desc+1)) != 0 )
    {
      gdImageSetTile(im, tmpim);
      color = gdTiled;
    }
  }
  else /* EVERYTHING failed! */
  {
    int totcolors = gdImageColorsTotal(im);
    if ( totcolors == 0 ) /* NO colors allocated at all! Allocate one and use it */
      color = gdImageColorAllocate(im, 255, 255, 255);
  }
  
  return color;
}

/*****************************************************************
** Translate font names to file names.
** This must be checked for failure (return of 0)
** The substitution table is currently static, but should be dynamic,
** read from a file. The file should be created through a program
** which reads the TTF files and writes the fonts it finds.
** Probably the subst table really ought to be a Tcl_HashTable,
** which would allow easy modification on the fly.
*****************************************************************/
struct gd_fontalias gd_subst_table[] =
{
  { "Times",       "Times New Roman" },
  { "Times Roman", "Times New Roman" },
  { "Helvetica",   "Arial" },
  { "Courier",     "Courier New" },
  { 0, 0 }
};

const char *gdFileFromFont(Tcl_Interp *interp, const char *font, double *ptsize, int debug)
{
  static char currentanswer[1024];
  int i;
  char **list;
  int count;
  char *fname = (char *)font;
  Tcl_DString dfont;
        
  int bold = 0;
  int italic = 0;
    
  Tcl_DStringInit(&dfont);
  currentanswer[0] = '\0';
 
  /* Shortcut */
  if ( subst_table == 0 )
    subst_table = gd_subst_table;
    
    
  if ( Tcl_SplitList(interp, (char *)font, &count, &list) == TCL_OK )
  {
      int j;
      
      Tcl_DStringAppend(&dfont, list[0], -1);
      
      /* Get the entire font spec -- with special handling of lowercase normal, bold, roman, and italic */
      /* Make an attempt to parse the result of using the Tk 8.2+ "font" command "font actual" result */
      for (j=1;list[j] != 0; j++)
      {
          double siz;
          char *strend;
          siz = strtod(list[j], &strend);
          if ( strend > list[j] ) /* It's a size */
              *ptsize = siz;
          else if ( strcmp(list[j], "underline") == 0 )
              ;
          else if ( strcmp(list[j], "overstrike") == 0 )
              ;
          else if ( strcmp(list[j], "-underline") == 0 )
              j++; /* Skip the next argument too */
          else if ( strcmp(list[j], "-overstrike") == 0 )
              j++; /* Skip the next argument too */
          else if ( list[j][0] == '-' )
              ; /* Part of a font specification -- ignore */
          else if ( strcmp(list[j], "normal") == 0 )
              bold = 0;
          else if ( strcmp(list[j], "roman") == 0 )
              italic = 0;
          else if ( strcmp(list[j], "bold") == 0 )
              bold = 1;
          else if ( strcmp(list[j], "italic") == 0 )
              italic = 1;
          else 
          {
              Tcl_DStringAppend(&dfont, " ", -1);
              Tcl_DStringAppend(&dfont, list[j], -1);
          }
      }
      fname = Tcl_DStringValue(&dfont);
      
      if ( debug )
        Tcl_AppendResult(interp, "{DEBUG: Constructed spec ", fname, " for font ", font, "}\n", 0);
      
      Tcl_Free((char *)list);
  }
    
  /* The subst table really needs to replace the font names, not provide the file names */
  /* First, do a substitution (case sensitive) */
  for (i = 0; subst_table[i].fontname; i++)
  {
    if ( strcmp(subst_table[i].fontname, fname) == 0 )
    {
      if ( debug )
          Tcl_AppendResult(interp, "{DEBUG: Using font ", subst_table[i].windows_fontname, " for font ", subst_table[i].fontname, "}\n", 0);
      Tcl_DStringFree(&dfont);
      Tcl_DStringInit(&dfont);
      Tcl_DStringAppend(&dfont, subst_table[i].windows_fontname, -1);  
      fname = Tcl_DStringValue(&dfont); /* Probably need to reset the pointer */
      break;
    }
  }
    
  /* Next, see if it's in the alias file */
  {
    const char *fn;
    if ( (fn = gdFontFile(fname,italic,bold)) != 0 )
    {
      if ( debug )
        Tcl_AppendResult(interp, "{DEBUG: Using file ", fn, " for font ", font, "}\n", 0);
        
      Tcl_DStringFree(&dfont);

      return fn;
    }
  }
    
  if ( debug )
    Tcl_AppendResult(interp, "{DEBUG: Could not find alias entry for ", fname, "}\n", 0);
    
  fname = 0;
    
#ifdef __WIN32__
  /* Under win32, we prefix the font by %WINDIR%\FONTS if no full pathname is present */
  /* We look for a two element list -- the first element is the font name,
  ** and the second is the font size.
  ** This ought to synchronize well with what is in GDI extension. This may require some
   * work with BWidgets or other extensions which integrate with Windows.
   * Perhaps the newer 8.x 'font' command could help in that regard...
  */
  {
      Tcl_DStringFree(&dfont);
      Tcl_DStringInit(&dfont);
      
      /* If the font is already specified to the file name level, don't do this step */
      if ( access(font, 04) == -1 && Tcl_SplitList(interp, (char *)font, &count, &list) == TCL_OK )
      {
          if ( count >= 1 )
          {
              if (list[0][0] != '\\' && list[0][0] != '/' && list[0][0] != '\0' && list[0][1] != ':' )
              {
                  const char *cp = getenv("WINDIR");
                  if ( cp )
                  {
                      Tcl_DStringAppend(&dfont, cp, -1);
                      Tcl_DStringAppend(&dfont, "\\fonts\\", -1);
                  }
              }
              font = list[0];
            }
          if ( count >= 2)
          {
              double siz;
              char *strend;
              siz = strtod(list[1], &strend);
              if ( strend > list[1] )
              {
                  *ptsize = siz;
              }
          }
          Tcl_Free( (char *)list);
      }
  }
#endif
    /* Otherwise, just add ".ttf" to the font to get the file */
    /* Note that the mapping between file names and windows names is not very easy
     * to use without actually loading every file. Perhaps this extension should 
     * create a .ini file or the equivalent to aid in this process--or at least
     * admit to an external alias table, that some tool could build. The alternative
     * is to "fix" the FreeType tools to do this mapping
     */
  Tcl_DStringAppend(&dfont, font, -1);
  if ( strstr(font, ".ttf") == 0 )
    Tcl_DStringAppend(&dfont, ".ttf", -1);
  if (debug)
      Tcl_AppendResult(interp, "{DEBUG: Picked font: ", Tcl_DStringValue(&dfont), "}\n", 0);

  strncpy(currentanswer, Tcl_DStringValue(&dfont), sizeof(currentanswer) );
  currentanswer[sizeof(currentanswer) - 1] = '\0';

  Tcl_DStringFree(&dfont);
    
  return currentanswer;
}

